# ruff: NoQA: D403, PIE796
import enum
from typing import final


class MemberType:
    """Custom data type with a simple API."""

    # this mangled attribute will never be shown on subclasses
    # even if :inherited-members: and :private-members: are set
    __slots__ = ('__data',)

    def __new__(cls, value):
        self = object.__new__(cls)
        self.__data = value
        return self

    def __str__(self):
        """inherited"""
        return self.__data

    def __repr__(self):
        return repr(self.__data)

    def __reduce__(self):
        # data types must be pickable, otherwise enum classes using this data
        # type will be forced to be non-pickable and have their __module__ set
        # to '<unknown>' instead of, for instance, '__main__'
        return self.__class__, (self.__data,)

    @final
    @property
    def dtype(self):
        """docstring"""
        return 'str'

    def isupper(self):
        """inherited"""
        return self.__data.isupper()


class EnumCls(enum.Enum):
    """this is enum class"""

    #: doc for val1
    val1 = 12
    val2 = 23  #: doc for val2
    val3 = 34
    """doc for val3"""
    val4 = 34

    def say_hello(self):
        """a method says hello to you."""

    @classmethod
    def say_goodbye(cls):
        """a classmethod says good-bye to you."""


class EnumClassWithDataType(MemberType, enum.Enum):
    """this is enum class"""

    x = 'x'

    def say_hello(self):
        """docstring"""

    @classmethod
    def say_goodbye(cls):
        """docstring"""


class ToUpperCase:  # not inheriting from enum.Enum
    @property
    def value(self):  # bypass enum.Enum.value
        """uppercased"""
        return str(self._value_).upper()  # type: ignore[attr-defined]


class Greeter:
    def say_hello(self):
        """inherited"""

    @classmethod
    def say_goodbye(cls):
        """inherited"""


class EnumClassWithMixinType(ToUpperCase, enum.Enum):
    """this is enum class"""

    x = 'x'

    def say_hello(self):
        """docstring"""

    @classmethod
    def say_goodbye(cls):
        """docstring"""


class EnumClassWithMixinTypeInherit(Greeter, ToUpperCase, enum.Enum):
    """this is enum class"""

    x = 'x'


class Overridden(enum.Enum):
    def override(self):
        """inherited"""
        return 1


class EnumClassWithMixinEnumType(Greeter, Overridden, enum.Enum):
    """this is enum class"""

    x = 'x'

    def override(self):
        """overridden"""
        return 2


class EnumClassWithMixinAndDataType(Greeter, ToUpperCase, MemberType, enum.Enum):
    """this is enum class"""

    x = 'x'

    def say_hello(self):
        """overridden"""

    @classmethod
    def say_goodbye(cls):
        """overridden"""

    def isupper(self):
        """overridden"""
        return False

    def __str__(self):
        """overridden"""
        return super().__str__()


class _ParentEnum(Greeter, Overridden, enum.Enum):
    """docstring"""


class EnumClassWithParentEnum(ToUpperCase, MemberType, _ParentEnum, enum.Enum):
    """this is enum class"""

    x = 'x'

    def isupper(self):
        """overridden"""
        return False

    def __str__(self):
        """overridden"""
        return super().__str__()


class _SunderMissingInNonEnumMixin:
    @classmethod
    def _missing_(cls, value):
        """inherited"""
        return super()._missing_(value)  # type: ignore[misc]


class _SunderMissingInEnumMixin(enum.Enum):
    @classmethod
    def _missing_(cls, value):
        """inherited"""
        return super()._missing_(value)


class _SunderMissingInDataType(MemberType):
    @classmethod
    def _missing_(cls, value):
        """inherited"""
        return super()._missing_(value)  # type: ignore[misc]


class EnumSunderMissingInNonEnumMixin(_SunderMissingInNonEnumMixin, enum.Enum):
    """this is enum class"""


class EnumSunderMissingInEnumMixin(_SunderMissingInEnumMixin, enum.Enum):
    """this is enum class"""


class EnumSunderMissingInDataType(_SunderMissingInDataType, enum.Enum):
    """this is enum class"""


class EnumSunderMissingInClass(enum.Enum):
    """this is enum class"""

    @classmethod
    def _missing_(cls, value):
        """docstring"""
        return super()._missing_(value)


class _NamePropertyInNonEnumMixin:
    @property
    def name(self):
        """inherited"""
        return super().name  # type: ignore[misc]


class _NamePropertyInEnumMixin(enum.Enum):
    @property
    def name(self):
        """inherited"""
        return super().name


class _NamePropertyInDataType(MemberType):
    @property
    def name(self):
        """inherited"""
        return super().name  # type: ignore[misc]


class EnumNamePropertyInNonEnumMixin(_NamePropertyInNonEnumMixin, enum.Enum):
    """this is enum class"""


class EnumNamePropertyInEnumMixin(_NamePropertyInEnumMixin, enum.Enum):
    """this is enum class"""


class EnumNamePropertyInDataType(_NamePropertyInDataType, enum.Enum):
    """this is enum class"""


class EnumNamePropertyInClass(enum.Enum):
    """this is enum class"""

    @property
    def name(self):
        """docstring"""
        return super().name



"""Test the HTML builder and check output against XPath."""

import re

import pytest

from tests.test_builders.xpath_util import check_xpath


def tail_check(check):
    rex = re.compile(check)

    def checker(nodes):
        for node in nodes:
            if node.tail and rex.search(node.tail):
                return True
        msg = f'{check!r} not found in tail of any nodes {nodes}'
        raise AssertionError(msg)
    return checker


@pytest.mark.parametrize(("fname", "path", "check"), [
    ('images.html', ".//img[@src='_images/img.png']", ''),
    ('images.html', ".//img[@src='_images/img1.png']", ''),
    ('images.html', ".//img[@src='_images/simg.png']", ''),
    ('images.html', ".//img[@src='_images/svgimg.svg']", ''),
    ('images.html', ".//a[@href='_sources/images.txt']", ''),

    ('subdir/images.html', ".//img[@src='../_images/img1.png']", ''),
    ('subdir/images.html', ".//img[@src='../_images/rimg.png']", ''),

    ('subdir/includes.html', ".//a[@class='reference download internal']", ''),
    ('subdir/includes.html', ".//img[@src='../_images/img.png']", ''),
    ('subdir/includes.html', ".//p", 'This is an include file.'),
    ('subdir/includes.html', ".//pre/span", 'line 1'),
    ('subdir/includes.html', ".//pre/span", 'line 2'),

    ('includes.html', ".//pre", 'Max Strauß'),
    ('includes.html', ".//a[@class='reference download internal']", ''),
    ('includes.html', ".//pre/span", '"quotes"'),
    ('includes.html', ".//pre/span", "'included'"),
    ('includes.html', ".//pre/span[@class='s2']", 'üöä'),
    ('includes.html', ".//div[@class='inc-pyobj1 highlight-text notranslate']//pre",
     r'^class Foo:\n    pass\n\s*$'),
    ('includes.html', ".//div[@class='inc-pyobj2 highlight-text notranslate']//pre",
     r'^    def baz\(\):\n        pass\n\s*$'),
    ('includes.html', ".//div[@class='inc-lines highlight-text notranslate']//pre",
     r'^class Foo:\n    pass\nclass Bar:\n$'),
    ('includes.html', ".//div[@class='inc-startend highlight-text notranslate']//pre",
     '^foo = "Including Unicode characters: üöä"\\n$'),
    ('includes.html', ".//div[@class='inc-preappend highlight-text notranslate']//pre",
     r'(?m)^START CODE$'),
    ('includes.html', ".//div[@class='inc-pyobj-dedent highlight-python notranslate']//span",
     r'def'),
    ('includes.html', ".//div[@class='inc-tab3 highlight-text notranslate']//pre",
     r'-| |-'),
    ('includes.html', ".//div[@class='inc-tab8 highlight-python notranslate']//pre/span",
     r'-|      |-'),

    ('autodoc.html', ".//dl[@class='py class']/dt[@id='autodoc_target.Class']", ''),
    ('autodoc.html', ".//dl[@class='py function']/dt[@id='autodoc_target.function']/em/span/span", r'\*\*'),
    ('autodoc.html', ".//dl[@class='py function']/dt[@id='autodoc_target.function']/em/span/span", r'kwds'),
    ('autodoc.html', ".//dd/p", r'Return spam\.'),

    ('extapi.html', ".//strong", 'from class: Bar'),

    ('markup.html', ".//title", 'set by title directive'),
    ('markup.html', ".//p/em", 'Section author: Georg Brandl'),
    ('markup.html', ".//p/em", 'Module author: Georg Brandl'),
    # created by the meta directive
    ('markup.html', ".//meta[@name='author'][@content='Me']", ''),
    ('markup.html', ".//meta[@name='keywords'][@content='docs, sphinx']", ''),
    # a label created by ``.. _label:``
    ('markup.html', ".//div[@id='label']", ''),
    # code with standard code blocks
    ('markup.html', ".//pre", '^some code$'),
    # an option list
    ('markup.html', ".//span[@class='option']", '--help'),
    # admonitions
    ('markup.html', ".//p[@class='admonition-title']", 'My Admonition'),
    ('markup.html', ".//div[@class='admonition note']/p", 'Note text.'),
    ('markup.html', ".//div[@class='admonition warning']/p", 'Warning text.'),
    # inline markup
    ('markup.html', ".//li/p/strong", r'^command\\n$'),
    ('markup.html', ".//li/p/strong", r'^program\\n$'),
    ('markup.html', ".//li/p/em", r'^dfn\\n$'),
    ('markup.html', ".//li/p/kbd", r'^kbd\\n$'),
    ('markup.html', ".//li/p/span", 'File \N{TRIANGULAR BULLET} Close'),
    ('markup.html', ".//li/p/code/span[@class='pre']", '^a/$'),
    ('markup.html', ".//li/p/code/em/span[@class='pre']", '^varpart$'),
    ('markup.html', ".//li/p/code/em/span[@class='pre']", '^i$'),
    ('markup.html', ".//a[@href='https://peps.python.org/pep-0008/']"
     "[@class='pep reference external']/strong", 'PEP 8'),
    ('markup.html', ".//a[@href='https://peps.python.org/pep-0008/']"
     "[@class='pep reference external']/strong",
     'Python Enhancement Proposal #8'),
    ('markup.html', ".//a[@href='https://datatracker.ietf.org/doc/html/rfc1.html']"
     "[@class='rfc reference external']/strong", 'RFC 1'),
    ('markup.html', ".//a[@href='https://datatracker.ietf.org/doc/html/rfc1.html']"
     "[@class='rfc reference external']/strong", 'Request for Comments #1'),
    ('markup.html', ".//a[@href='objects.html#envvar-HOME']"
     "[@class='reference internal']/code/span[@class='pre']", 'HOME'),
    ('markup.html', ".//a[@href='#with']"
     "[@class='reference internal']/code/span[@class='pre']", '^with$'),
    ('markup.html', ".//a[@href='#grammar-token-try_stmt']"
     "[@class='reference internal']/code/span", '^statement$'),
    ('markup.html', ".//a[@href='#some-label'][@class='reference internal']/span", '^here$'),
    ('markup.html', ".//a[@href='#some-label'][@class='reference internal']/span", '^there$'),
    ('markup.html', ".//a[@href='subdir/includes.html']"
     "[@class='reference internal']/span", 'Including in subdir'),
    ('markup.html', ".//a[@href='objects.html#cmdoption-python-c']"
     "[@class='reference internal']/code/span[@class='pre']", '-c'),
    # abbreviations
    ('markup.html', ".//abbr[@title='abbreviation']", '^abbr$'),
    # version stuff
    ('markup.html', ".//div[@class='versionadded']/p/span", 'Added in version 0.6: '),
    ('markup.html', ".//div[@class='versionadded']/p/span",
     tail_check('First paragraph of versionadded')),
    ('markup.html', ".//div[@class='versionchanged']/p/span",
     tail_check('First paragraph of versionchanged')),
    ('markup.html', ".//div[@class='versionchanged']/p",
     'Second paragraph of versionchanged'),
    ('markup.html', ".//div[@class='versionremoved']/p/span", 'Removed in version 0.6: '),
    # footnote reference
    ('markup.html', ".//a[@class='footnote-reference brackets']", r'1'),
    # created by reference lookup
    ('markup.html', ".//a[@href='index.html#ref1']", ''),
    # ``seealso`` directive
    ('markup.html', ".//div/p[@class='admonition-title']", 'See also'),
    # a ``hlist`` directive
    ('markup.html', ".//table[@class='hlist']/tr/td/ul/li/p", '^This$'),
    # a ``centered`` directive
    ('markup.html', ".//p[@class='centered']/strong", 'LICENSE'),
    # a glossary
    ('markup.html', ".//dl/dt[@id='term-boson']", 'boson'),
    ('markup.html', ".//dl/dt[@id='term-boson']/a", '¶'),
    # a production list
    ('markup.html', ".//pre/strong", 'try_stmt'),
    ('markup.html', ".//pre/a[@href='#grammar-token-try1_stmt']/code/span", 'try1_stmt'),
    # tests for ``only`` directive
    ('markup.html', ".//p", 'A global substitution!'),
    ('markup.html', ".//p", 'In HTML.'),
    ('markup.html', ".//p", 'In both.'),
    ('markup.html', ".//p", 'Always present'),
    # tests for ``any`` role
    ('markup.html', ".//a[@href='#with']/span", 'headings'),
    ('markup.html', ".//a[@href='objects.html#func_without_body']/code/span", 'objects'),
    # tests for numeric labels
    ('markup.html', ".//a[@href='#id1'][@class='reference internal']/span", 'Testing various markup'),
    # tests for smartypants
    ('markup.html', ".//li/p", 'Smart “quotes” in English ‘text’.'),
    ('markup.html', ".//li/p", 'Smart — long and – short dashes.'),
    ('markup.html', ".//li/p", 'Ellipsis…'),
    ('markup.html', ".//li/p/code/span[@class='pre']", 'foo--"bar"...'),
    ('markup.html', ".//p", 'Этот «абзац» должен использовать „русские“ кавычки.'),
    ('markup.html', ".//p", 'Il dit : « C’est “super” ! »'),

    ('objects.html', ".//dt[@id='mod.Cls.meth1']", ''),
    ('objects.html', ".//dt[@id='errmod.Error']", ''),
    ('objects.html', ".//dt/span[@class='sig-name descname']/span[@class='pre']", r'long\(parameter,'),
    ('objects.html', ".//dt/span[@class='sig-name descname']/span[@class='pre']", r'list\)'),
    ('objects.html', ".//dt/span[@class='sig-name descname']/span[@class='pre']", 'another'),
    ('objects.html', ".//dt/span[@class='sig-name descname']/span[@class='pre']", 'one'),
    ('objects.html', ".//a[@href='#mod.Cls'][@class='reference internal']", ''),
    ('objects.html', ".//dl[@class='std userdesc']", ''),
    ('objects.html', ".//dt[@id='userdesc-myobj']", ''),
    ('objects.html', ".//a[@href='#userdesc-myobj'][@class='reference internal']", ''),
    # docfields
    ('objects.html', ".//a[@class='reference internal'][@href='#TimeInt']/em", 'TimeInt'),
    ('objects.html', ".//a[@class='reference internal'][@href='#Time']", 'Time'),
    ('objects.html', ".//a[@class='reference internal'][@href='#errmod.Error']/strong", 'Error'),
    # C references
    ('objects.html', ".//span[@class='pre']", 'CFunction()'),
    ('objects.html', ".//a[@href='#c.Sphinx_DoSomething']", ''),
    ('objects.html', ".//a[@href='#c.SphinxStruct.member']", ''),
    ('objects.html', ".//a[@href='#c.SPHINX_USE_PYTHON']", ''),
    ('objects.html', ".//a[@href='#c.SphinxType']", ''),
    ('objects.html', ".//a[@href='#c.sphinx_global']", ''),
    # test global TOC created by toctree()
    ('objects.html', ".//ul[@class='current']/li[@class='toctree-l1 current']/a[@href='#']",
     'Testing object descriptions'),
    ('objects.html', ".//li[@class='toctree-l1']/a[@href='markup.html']",
     'Testing various markup'),
    # test unknown field names
    ('objects.html', ".//dt[@class='field-odd']", 'Field_name'),
    ('objects.html', ".//dt[@class='field-even']", 'Field_name all lower'),
    ('objects.html', ".//dt[@class='field-odd']", 'FIELD_NAME'),
    ('objects.html', ".//dt[@class='field-even']", 'FIELD_NAME ALL CAPS'),
    ('objects.html', ".//dt[@class='field-odd']", 'Field_Name'),
    ('objects.html', ".//dt[@class='field-even']", 'Field_Name All Word Caps'),
    # ('objects.html', ".//dt[@class='field-odd']", 'Field_name'), (duplicate)
    ('objects.html', ".//dt[@class='field-even']", 'Field_name First word cap'),
    ('objects.html', ".//dt[@class='field-odd']", 'FIELd_name'),
    ('objects.html', ".//dt[@class='field-even']", 'FIELd_name PARTial caps'),
    # custom sidebar
    ('objects.html', ".//h4", 'Custom sidebar'),
    # docfields
    ('objects.html', ".//dd[@class='field-odd']/p/strong", '^moo$'),
    ('objects.html', ".//dd[@class='field-odd']/p/strong", tail_check(r'\(Moo\) .* Moo')),
    ('objects.html', ".//dd[@class='field-odd']/ul/li/p/strong", '^hour$'),
    ('objects.html', ".//dd[@class='field-odd']/ul/li/p/em", '^DuplicateType$'),
    ('objects.html', ".//dd[@class='field-odd']/ul/li/p/em", tail_check(r'.* Some parameter')),
    # others
    ('objects.html', ".//a[@class='reference internal'][@href='#cmdoption-perl-arg-p']/code/span",
     'perl'),
    ('objects.html', ".//a[@class='reference internal'][@href='#cmdoption-perl-arg-p']/code/span",
     '\\+p'),
    ('objects.html', ".//a[@class='reference internal'][@href='#cmdoption-perl-ObjC']/code/span",
     '--ObjC\\+\\+'),
    ('objects.html', ".//a[@class='reference internal'][@href='#cmdoption-perl-plugin.option']/code/span",
     '--plugin.option'),
    ('objects.html', ".//a[@class='reference internal'][@href='#cmdoption-perl-arg-create-auth-token']"
     "/code/span",
     'create-auth-token'),
    ('objects.html', ".//a[@class='reference internal'][@href='#cmdoption-perl-arg-arg']/code/span",
     'arg'),
    ('objects.html', ".//a[@class='reference internal'][@href='#cmdoption-perl-j']/code/span",
     '-j'),
    ('objects.html', ".//a[@class='reference internal'][@href='#cmdoption-hg-arg-commit']/code/span",
     'hg'),
    ('objects.html', ".//a[@class='reference internal'][@href='#cmdoption-hg-arg-commit']/code/span",
     'commit'),
    ('objects.html', ".//a[@class='reference internal'][@href='#cmdoption-git-commit-p']/code/span",
     'git'),
    ('objects.html', ".//a[@class='reference internal'][@href='#cmdoption-git-commit-p']/code/span",
     'commit'),
    ('objects.html', ".//a[@class='reference internal'][@href='#cmdoption-git-commit-p']/code/span",
     '-p'),

    ('index.html', ".//meta[@name='hc'][@content='hcval']", ''),
    ('index.html', ".//meta[@name='hc_co'][@content='hcval_co']", ''),
    ('index.html', ".//li[@class='toctree-l1']/a", 'Testing various markup'),
    ('index.html', ".//li[@class='toctree-l2']/a", 'Inline markup'),
    ('index.html', ".//title", 'Sphinx <Tests>'),
    ('index.html', ".//div[@class='footer']", 'copyright text credits'),
    ('index.html', ".//a[@href='https://python.org/']"
     "[@class='reference external']", ''),
    ('index.html', ".//li/p/a[@href='genindex.html']/span", 'Index'),
    ('index.html', ".//li/p/a[@href='py-modindex.html']/span", 'Module Index'),
    # custom sidebar only for contents
    ('index.html', ".//h4", 'Contents sidebar'),
    # custom JavaScript
    ('index.html', ".//script[@src='file://moo.js']", ''),
    # URL in contents
    ('index.html', ".//a[@class='reference external'][@href='https://sphinx-doc.org/']",
     'https://sphinx-doc.org/'),
    ('index.html', ".//a[@class='reference external'][@href='https://sphinx-doc.org/latest/']",
     'Latest reference'),
    # Indirect hyperlink targets across files
    ('index.html', ".//a[@href='markup.html#some-label'][@class='reference internal']/span",
     '^indirect hyperref$'),

    ('bom.html', ".//title", " File with UTF-8 BOM"),

    ('extensions.html', ".//a[@href='https://python.org/dev/']", "https://python.org/dev/"),
    ('extensions.html', ".//a[@href='https://bugs.python.org/issue1000']", "issue 1000"),
    ('extensions.html', ".//a[@href='https://bugs.python.org/issue1042']", "explicit caption"),

    # index entries
    ('genindex.html', ".//a/strong", "Main"),
    ('genindex.html', ".//a/strong", "[1]"),
    ('genindex.html', ".//a/strong", "Other"),
    ('genindex.html', ".//a", "entry"),
    ('genindex.html', ".//li/a", "double"),

    ('otherext.html', ".//h1", "Generated section"),
    ('otherext.html', ".//a[@href='_sources/otherext.foo.txt']", ''),

    ('search.html', ".//meta[@name='robots'][@content='noindex']", ''),
])
@pytest.mark.sphinx('html', tags=['testtag'],
                    confoverrides={'html_context.hckey_co': 'hcval_co'})
@pytest.mark.test_params(shared_result='test_build_html_output')
def test_html5_output(app, cached_etree_parse, fname, path, check):
    app.build()
    check_xpath(cached_etree_parse(app.outdir / fname), fname, path, check)



"""Input/Output files"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

import docutils
from docutils import nodes
from docutils.core import Publisher
from docutils.io import FileInput, Input, NullOutput
from docutils.readers import standalone
from docutils.transforms.references import DanglingReferences
from docutils.writers import UnfilteredWriter

from sphinx import addnodes
from sphinx.transforms import AutoIndexUpgrader, DoctreeReadEvent, SphinxTransformer
from sphinx.transforms.i18n import (
    Locale,
    PreserveTranslatableMessages,
    RemoveTranslatableInline,
)
from sphinx.transforms.references import SphinxDomains
from sphinx.util import logging
from sphinx.util.docutils import LoggingReporter
from sphinx.versioning import UIDTransform

if TYPE_CHECKING:
    from docutils.frontend import Values
    from docutils.parsers import Parser
    from docutils.transforms import Transform

    from sphinx.application import Sphinx
    from sphinx.environment import BuildEnvironment


logger = logging.getLogger(__name__)


class SphinxBaseReader(standalone.Reader):
    """
    A base class of readers for Sphinx.

    This replaces reporter by Sphinx's on generating document.
    """

    transforms: list[type[Transform]] = []

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        from sphinx.application import Sphinx

        if len(args) > 0 and isinstance(args[0], Sphinx):
            self._app = args[0]
            self._env = self._app.env
            args = args[1:]

        super().__init__(*args, **kwargs)

    def setup(self, app: Sphinx) -> None:
        self._app = app  # hold application object only for compatibility
        self._env = app.env

    def get_transforms(self) -> list[type[Transform]]:
        transforms = super().get_transforms() + self.transforms

        # remove transforms which is not needed for Sphinx
        unused = [DanglingReferences]
        for transform in unused:
            if transform in transforms:
                transforms.remove(transform)

        return transforms

    def new_document(self) -> nodes.document:
        """
        Creates a new document object which has a special reporter object good
        for logging.
        """
        document = super().new_document()
        document.__class__ = addnodes.document  # replace the class with patched version

        # substitute transformer
        document.transformer = SphinxTransformer(document)
        document.transformer.set_environment(self.settings.env)

        # substitute reporter
        reporter = document.reporter
        document.reporter = LoggingReporter.from_reporter(reporter)

        return document


class SphinxStandaloneReader(SphinxBaseReader):
    """
    A basic document reader for Sphinx.
    """

    def setup(self, app: Sphinx) -> None:
        self.transforms = self.transforms + app.registry.get_transforms()
        super().setup(app)

    def read(self, source: Input, parser: Parser, settings: Values) -> nodes.document:  # type: ignore[type-arg]
        self.source = source
        if not self.parser:  # type: ignore[has-type]
            self.parser = parser
        self.settings = settings
        self.input = self.read_source(settings.env)
        self.parse()
        return self.document

    def read_source(self, env: BuildEnvironment) -> str:
        """Read content from source and do post-process."""
        content = self.source.read()

        # emit "source-read" event
        arg = [content]
        env.events.emit('source-read', env.docname, arg)
        return arg[0]


class SphinxI18nReader(SphinxBaseReader):
    """
    A document reader for i18n.

    This returns the source line number of original text as current source line number
    to let users know where the error happened.
    Because the translated texts are partial and they don't have correct line numbers.
    """

    def setup(self, app: Sphinx) -> None:
        super().setup(app)

        self.transforms = self.transforms + app.registry.get_transforms()
        unused = [
            PreserveTranslatableMessages,
            Locale,
            RemoveTranslatableInline,
            AutoIndexUpgrader,
            SphinxDomains,
            DoctreeReadEvent,
            UIDTransform,
        ]
        for transform in unused:
            if transform in self.transforms:
                self.transforms.remove(transform)


class SphinxDummyWriter(UnfilteredWriter):
    """Dummy writer module used for generating doctree."""

    supported = ('html',)  # needed to keep "meta" nodes

    def translate(self) -> None:
        pass


def SphinxDummySourceClass(source: Any, *args: Any, **kwargs: Any) -> Any:
    """Bypass source object as is to cheat Publisher."""
    return source


class SphinxFileInput(FileInput):
    """A basic FileInput for Sphinx."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        kwargs['error_handler'] = 'sphinx'
        super().__init__(*args, **kwargs)


def create_publisher(app: Sphinx, filetype: str) -> Publisher:
    reader = SphinxStandaloneReader()
    reader.setup(app)

    parser = app.registry.create_source_parser(app, filetype)
    if parser.__class__.__name__ == 'CommonMarkParser' and parser.settings_spec == ():
        # a workaround for recommonmark
        #   If recommonmark.AutoStrictify is enabled, the parser invokes reST parser
        #   internally.  But recommonmark-0.4.0 does not provide settings_spec for reST
        #   parser.  As a workaround, this copies settings_spec for RSTParser to the
        #   CommonMarkParser.
        from docutils.parsers.rst import Parser as RSTParser

        parser.settings_spec = RSTParser.settings_spec  # type: ignore[misc]

    pub = Publisher(
        reader=reader,
        parser=parser,
        writer=SphinxDummyWriter(),
        source_class=SphinxFileInput,
        destination=NullOutput(),
    )
    # Propagate exceptions by default when used programmatically:
    defaults = {'traceback': True, **app.env.settings}
    # Set default settings
    if docutils.__version_info__[:2] >= (0, 19):
        pub.get_settings(**defaults)
    else:
        pub.settings = pub.setup_option_parser(**defaults).get_default_values()
    return pub



"""Tests the Python Domain"""

from __future__ import annotations

import pytest

from sphinx import addnodes
from sphinx.addnodes import (
    desc,
    desc_addname,
    desc_annotation,
    desc_content,
    desc_name,
    desc_sig_space,
    desc_signature,
)
from sphinx.testing import restructuredtext
from sphinx.testing.util import assert_node


@pytest.mark.sphinx('html', testroot='domain-py', freshenv=True)
def test_domain_py_canonical(app, status, warning):
    app.build(force_all=True)

    content = (app.outdir / 'canonical.html').read_text(encoding='utf8')
    assert ('<a class="reference internal" href="#canonical.Foo" title="canonical.Foo">'
            '<code class="xref py py-class docutils literal notranslate">'
            '<span class="pre">Foo</span></code></a>' in content)
    assert warning.getvalue() == ''


def test_canonical(app):
    text = (".. py:class:: io.StringIO\n"
            "   :canonical: _io.StringIO")
    domain = app.env.get_domain('py')
    doctree = restructuredtext.parse(app, text)
    assert_node(doctree, (addnodes.index,
                          [desc, ([desc_signature, ([desc_annotation, ("class", desc_sig_space)],
                                                    [desc_addname, "io."],
                                                    [desc_name, "StringIO"])],
                                  desc_content)]))
    assert 'io.StringIO' in domain.objects
    assert domain.objects['io.StringIO'] == ('index', 'io.StringIO', 'class', False)
    assert domain.objects['_io.StringIO'] == ('index', 'io.StringIO', 'class', True)


def test_canonical_definition_overrides(app, warning):
    text = (".. py:class:: io.StringIO\n"
            "   :canonical: _io.StringIO\n"
            ".. py:class:: _io.StringIO\n")
    restructuredtext.parse(app, text)
    assert warning.getvalue() == ""

    domain = app.env.get_domain('py')
    assert domain.objects['_io.StringIO'] == ('index', 'id0', 'class', False)


def test_canonical_definition_skip(app, warning):
    text = (".. py:class:: _io.StringIO\n"
            ".. py:class:: io.StringIO\n"
            "   :canonical: _io.StringIO\n")

    restructuredtext.parse(app, text)
    assert warning.getvalue() == ""

    domain = app.env.get_domain('py')
    assert domain.objects['_io.StringIO'] == ('index', 'io.StringIO', 'class', False)


def test_canonical_duplicated(app, warning):
    text = (".. py:class:: mypackage.StringIO\n"
            "   :canonical: _io.StringIO\n"
            ".. py:class:: io.StringIO\n"
            "   :canonical: _io.StringIO\n")

    restructuredtext.parse(app, text)
    assert warning.getvalue() != ""



"""Single HTML builders."""

from __future__ import annotations

from os import path
from typing import TYPE_CHECKING, Any

from docutils import nodes

from sphinx.builders.html import StandaloneHTMLBuilder
from sphinx.environment.adapters.toctree import global_toctree_for_doc
from sphinx.locale import __
from sphinx.util import logging
from sphinx.util.console import darkgreen
from sphinx.util.display import progress_message
from sphinx.util.nodes import inline_all_toctrees

if TYPE_CHECKING:
    from docutils.nodes import Node

    from sphinx.application import Sphinx
    from sphinx.util.typing import ExtensionMetadata

logger = logging.getLogger(__name__)


class SingleFileHTMLBuilder(StandaloneHTMLBuilder):
    """
    A StandaloneHTMLBuilder subclass that puts the whole document tree on one
    HTML page.
    """

    name = 'singlehtml'
    epilog = __('The HTML page is in %(outdir)s.')

    copysource = False

    def get_outdated_docs(self) -> str | list[str]:  # type: ignore[override]
        return 'all documents'

    def get_target_uri(self, docname: str, typ: str | None = None) -> str:
        if docname in self.env.all_docs:
            # all references are on the same page...
            return '#document-' + docname
        else:
            # chances are this is a html_additional_page
            return docname + self.out_suffix

    def get_relative_uri(self, from_: str, to: str, typ: str | None = None) -> str:
        # ignore source
        return self.get_target_uri(to, typ)

    def fix_refuris(self, tree: Node) -> None:
        # fix refuris with double anchor
        fname = self.config.root_doc + self.out_suffix
        for refnode in tree.findall(nodes.reference):
            if 'refuri' not in refnode:
                continue
            refuri = refnode['refuri']
            hashindex = refuri.find('#')
            if hashindex < 0:
                continue
            hashindex = refuri.find('#', hashindex + 1)
            if hashindex >= 0:
                refnode['refuri'] = fname + refuri[hashindex:]

    def _get_local_toctree(self, docname: str, collapse: bool = True, **kwargs: Any) -> str:
        if isinstance(includehidden := kwargs.get('includehidden'), str):
            if includehidden.lower() == 'false':
                kwargs['includehidden'] = False
            elif includehidden.lower() == 'true':
                kwargs['includehidden'] = True
        if kwargs.get('maxdepth') == '':
            kwargs.pop('maxdepth')
        toctree = global_toctree_for_doc(self.env, docname, self, collapse=collapse, **kwargs)
        if toctree is not None:
            self.fix_refuris(toctree)
        return self.render_partial(toctree)['fragment']

    def assemble_doctree(self) -> nodes.document:
        master = self.config.root_doc
        tree = self.env.get_doctree(master)
        tree = inline_all_toctrees(self, set(), master, tree, darkgreen, [master])
        tree['docname'] = master
        self.env.resolve_references(tree, master, self)
        self.fix_refuris(tree)
        return tree

    def assemble_toc_secnumbers(self) -> dict[str, dict[str, tuple[int, ...]]]:
        # Assemble toc_secnumbers to resolve section numbers on SingleHTML.
        # Merge all secnumbers to single secnumber.
        #
        # Note: current Sphinx has refid confliction in singlehtml mode.
        #       To avoid the problem, it replaces key of secnumbers to
        #       tuple of docname and refid.
        #
        #       There are related codes in inline_all_toctres() and
        #       HTMLTranslter#add_secnumber().
        new_secnumbers: dict[str, tuple[int, ...]] = {}
        for docname, secnums in self.env.toc_secnumbers.items():
            for id, secnum in secnums.items():
                alias = f"{docname}/{id}"
                new_secnumbers[alias] = secnum

        return {self.config.root_doc: new_secnumbers}

    def assemble_toc_fignumbers(self) -> dict[str, dict[str, dict[str, tuple[int, ...]]]]:
        # Assemble toc_fignumbers to resolve figure numbers on SingleHTML.
        # Merge all fignumbers to single fignumber.
        #
        # Note: current Sphinx has refid confliction in singlehtml mode.
        #       To avoid the problem, it replaces key of secnumbers to
        #       tuple of docname and refid.
        #
        #       There are related codes in inline_all_toctres() and
        #       HTMLTranslter#add_fignumber().
        new_fignumbers: dict[str, dict[str, tuple[int, ...]]] = {}
        # {'foo': {'figure': {'id2': (2,), 'id1': (1,)}}, 'bar': {'figure': {'id1': (3,)}}}
        for docname, fignumlist in self.env.toc_fignumbers.items():
            for figtype, fignums in fignumlist.items():
                alias = f"{docname}/{figtype}"
                new_fignumbers.setdefault(alias, {})
                for id, fignum in fignums.items():
                    new_fignumbers[alias][id] = fignum

        return {self.config.root_doc: new_fignumbers}

    def get_doc_context(self, docname: str, body: str, metatags: str) -> dict[str, Any]:
        # no relation links...
        toctree = global_toctree_for_doc(self.env, self.config.root_doc, self, collapse=False)
        # if there is no toctree, toc is None
        if toctree:
            self.fix_refuris(toctree)
            toc = self.render_partial(toctree)['fragment']
            display_toc = True
        else:
            toc = ''
            display_toc = False
        return {
            'parents': [],
            'prev': None,
            'next': None,
            'docstitle': None,
            'title': self.config.html_title,
            'meta': None,
            'body': body,
            'metatags': metatags,
            'rellinks': [],
            'sourcename': '',
            'toc': toc,
            'display_toc': display_toc,
        }

    def write(self, *ignored: Any) -> None:
        docnames = self.env.all_docs

        with progress_message(__('preparing documents')):
            self.prepare_writing(docnames)  # type: ignore[arg-type]

        with progress_message(__('assembling single document')):
            doctree = self.assemble_doctree()
            self.env.toc_secnumbers = self.assemble_toc_secnumbers()
            self.env.toc_fignumbers = self.assemble_toc_fignumbers()

        with progress_message(__('writing')):
            self.write_doc_serialized(self.config.root_doc, doctree)
            self.write_doc(self.config.root_doc, doctree)

    def finish(self) -> None:
        self.write_additional_files()
        self.copy_image_files()
        self.copy_download_files()
        self.copy_static_files()
        self.copy_extra_files()
        self.write_buildinfo()
        self.dump_inventory()

    @progress_message(__('writing additional files'))
    def write_additional_files(self) -> None:
        # no indices or search pages are supported

        # additional pages from conf.py
        for pagename, template in self.config.html_additional_pages.items():
            logger.info(' ' + pagename, nonl=True)
            self.handle_page(pagename, {}, template)

        if self.config.html_use_opensearch:
            logger.info(' opensearch', nonl=True)
            fn = path.join(self.outdir, '_static', 'opensearch.xml')
            self.handle_page('opensearch', {}, 'opensearch.xml', outfilename=fn)


def setup(app: Sphinx) -> ExtensionMetadata:
    app.setup_extension('sphinx.builders.html')

    app.add_builder(SingleFileHTMLBuilder)
    app.add_config_value('singlehtml_sidebars', lambda self: self.html_sidebars, 'html')

    return {
        'version': 'builtin',
        'parallel_read_safe': True,
        'parallel_write_safe': True,
    }



"""Test the HTML builder and check output against XPath."""

import os
import posixpath
import re

import pytest

from sphinx.builders.html import validate_html_extra_path, validate_html_static_path
from sphinx.deprecation import RemovedInSphinx80Warning
from sphinx.errors import ConfigError
from sphinx.util.console import strip_colors
from sphinx.util.inventory import InventoryFile

from tests.test_builders.xpath_data import FIGURE_CAPTION
from tests.test_builders.xpath_util import check_xpath


def test_html4_error(make_app, tmp_path):
    (tmp_path / 'conf.py').write_text('', encoding='utf-8')
    with pytest.raises(
        ConfigError,
        match='HTML 4 is no longer supported by Sphinx',
    ):
        make_app(
            buildername='html',
            srcdir=tmp_path,
            confoverrides={'html4_writer': True},
        )


@pytest.mark.parametrize(("fname", "path", "check"), [
    ('index.html', ".//div[@class='citation']/span", r'Ref1'),
    ('index.html', ".//div[@class='citation']/span", r'Ref_1'),

    ('footnote.html', ".//a[@class='footnote-reference brackets'][@href='#id9'][@id='id1']", r"1"),
    ('footnote.html', ".//a[@class='footnote-reference brackets'][@href='#id10'][@id='id2']", r"2"),
    ('footnote.html', ".//a[@class='footnote-reference brackets'][@href='#foo'][@id='id3']", r"3"),
    ('footnote.html', ".//a[@class='reference internal'][@href='#bar'][@id='id4']/span", r"\[bar\]"),
    ('footnote.html', ".//a[@class='reference internal'][@href='#baz-qux'][@id='id5']/span", r"\[baz_qux\]"),
    ('footnote.html', ".//a[@class='footnote-reference brackets'][@href='#id11'][@id='id6']", r"4"),
    ('footnote.html', ".//a[@class='footnote-reference brackets'][@href='#id12'][@id='id7']", r"5"),
    ('footnote.html', ".//aside[@class='footnote brackets']/span/a[@href='#id1']", r"1"),
    ('footnote.html', ".//aside[@class='footnote brackets']/span/a[@href='#id2']", r"2"),
    ('footnote.html', ".//aside[@class='footnote brackets']/span/a[@href='#id3']", r"3"),
    ('footnote.html', ".//div[@class='citation']/span/a[@href='#id4']", r"bar"),
    ('footnote.html', ".//div[@class='citation']/span/a[@href='#id5']", r"baz_qux"),
    ('footnote.html', ".//aside[@class='footnote brackets']/span/a[@href='#id6']", r"4"),
    ('footnote.html', ".//aside[@class='footnote brackets']/span/a[@href='#id7']", r"5"),
    ('footnote.html', ".//aside[@class='footnote brackets']/span/a[@href='#id8']", r"6"),
])
@pytest.mark.sphinx('html')
@pytest.mark.test_params(shared_result='test_build_html_output_docutils18')
def test_docutils_output(app, cached_etree_parse, fname, path, check):
    app.build()
    check_xpath(cached_etree_parse(app.outdir / fname), fname, path, check)


@pytest.mark.sphinx('html', parallel=2)
def test_html_parallel(app):
    app.build()


@pytest.mark.sphinx('html', testroot='build-html-translator')
def test_html_translator(app):
    app.build()
    assert app.builder.docwriter.visitor.depart_with_node == 10


@pytest.mark.parametrize("expect", [
    (FIGURE_CAPTION + "//span[@class='caption-number']", "Fig. 1", True),
    (FIGURE_CAPTION + "//span[@class='caption-number']", "Fig. 2", True),
    (FIGURE_CAPTION + "//span[@class='caption-number']", "Fig. 3", True),
    (".//div//span[@class='caption-number']", "No.1 ", True),
    (".//div//span[@class='caption-number']", "No.2 ", True),
    (".//li/p/a/span", 'Fig. 1', True),
    (".//li/p/a/span", 'Fig. 2', True),
    (".//li/p/a/span", 'Fig. 3', True),
    (".//li/p/a/span", 'No.1', True),
    (".//li/p/a/span", 'No.2', True),
])
@pytest.mark.sphinx('html', testroot='add_enumerable_node',
                    srcdir='test_enumerable_node')
def test_enumerable_node(app, cached_etree_parse, expect):
    app.build()
    check_xpath(cached_etree_parse(app.outdir / 'index.html'), 'index.html', *expect)


@pytest.mark.sphinx('html', testroot='basic', confoverrides={'html_copy_source': False})
def test_html_copy_source(app):
    app.build(force_all=True)
    assert not (app.outdir / '_sources' / 'index.rst.txt').exists()


@pytest.mark.sphinx('html', testroot='basic', confoverrides={'html_sourcelink_suffix': '.txt'})
def test_html_sourcelink_suffix(app):
    app.build(force_all=True)
    assert (app.outdir / '_sources' / 'index.rst.txt').exists()


@pytest.mark.sphinx('html', testroot='basic', confoverrides={'html_sourcelink_suffix': '.rst'})
def test_html_sourcelink_suffix_same(app):
    app.build(force_all=True)
    assert (app.outdir / '_sources' / 'index.rst').exists()


@pytest.mark.sphinx('html', testroot='basic', confoverrides={'html_sourcelink_suffix': ''})
def test_html_sourcelink_suffix_empty(app):
    app.build(force_all=True)
    assert (app.outdir / '_sources' / 'index.rst').exists()


@pytest.mark.sphinx('html', testroot='html_entity')
def test_html_entity(app):
    app.build(force_all=True)
    valid_entities = {'amp', 'lt', 'gt', 'quot', 'apos'}
    content = (app.outdir / 'index.html').read_text(encoding='utf8')
    for entity in re.findall(r'&([a-z]+);', content, re.MULTILINE):
        assert entity not in valid_entities


@pytest.mark.sphinx('html', testroot='basic')
def test_html_inventory(app):
    app.build(force_all=True)

    with app.outdir.joinpath('objects.inv').open('rb') as f:
        invdata = InventoryFile.load(f, 'https://www.google.com', posixpath.join)

    assert set(invdata.keys()) == {'std:label', 'std:doc'}
    assert set(invdata['std:label'].keys()) == {'modindex',
                                                'py-modindex',
                                                'genindex',
                                                'search'}
    assert invdata['std:label']['modindex'] == ('Python',
                                                '',
                                                'https://www.google.com/py-modindex.html',
                                                'Module Index')
    assert invdata['std:label']['py-modindex'] == ('Python',
                                                   '',
                                                   'https://www.google.com/py-modindex.html',
                                                   'Python Module Index')
    assert invdata['std:label']['genindex'] == ('Python',
                                                '',
                                                'https://www.google.com/genindex.html',
                                                'Index')
    assert invdata['std:label']['search'] == ('Python',
                                              '',
                                              'https://www.google.com/search.html',
                                              'Search Page')
    assert set(invdata['std:doc'].keys()) == {'index'}
    assert invdata['std:doc']['index'] == ('Python',
                                           '',
                                           'https://www.google.com/index.html',
                                           'The basic Sphinx documentation for testing')


@pytest.mark.sphinx('html', testroot='images', confoverrides={'html_sourcelink_suffix': ''})
def test_html_anchor_for_figure(app):
    app.build(force_all=True)
    content = (app.outdir / 'index.html').read_text(encoding='utf8')
    assert ('<figcaption>\n<p><span class="caption-text">The caption of pic</span>'
            '<a class="headerlink" href="#id1" title="Link to this image">¶</a></p>\n</figcaption>'
            in content)


@pytest.mark.sphinx('html', testroot='directives-raw')
def test_html_raw_directive(app, status, warning):
    app.build(force_all=True)
    result = (app.outdir / 'index.html').read_text(encoding='utf8')

    # standard case
    assert 'standalone raw directive (HTML)' in result
    assert 'standalone raw directive (LaTeX)' not in result

    # with substitution
    assert '<p>HTML: abc def ghi</p>' in result
    assert '<p>LaTeX: abc  ghi</p>' in result


@pytest.mark.parametrize("expect", [
    (".//link[@href='_static/persistent.css']"
     "[@rel='stylesheet']", '', True),
    (".//link[@href='_static/default.css']"
     "[@rel='stylesheet']"
     "[@title='Default']", '', True),
    (".//link[@href='_static/alternate1.css']"
     "[@rel='alternate stylesheet']"
     "[@title='Alternate']", '', True),
    (".//link[@href='_static/alternate2.css']"
     "[@rel='alternate stylesheet']", '', True),
    (".//link[@href='_static/more_persistent.css']"
     "[@rel='stylesheet']", '', True),
    (".//link[@href='_static/more_default.css']"
     "[@rel='stylesheet']"
     "[@title='Default']", '', True),
    (".//link[@href='_static/more_alternate1.css']"
     "[@rel='alternate stylesheet']"
     "[@title='Alternate']", '', True),
    (".//link[@href='_static/more_alternate2.css']"
     "[@rel='alternate stylesheet']", '', True),
])
@pytest.mark.sphinx('html', testroot='stylesheets')
def test_alternate_stylesheets(app, cached_etree_parse, expect):
    app.build()
    check_xpath(cached_etree_parse(app.outdir / 'index.html'), 'index.html', *expect)


@pytest.mark.sphinx('html', testroot='html_style')
def test_html_style(app, status, warning):
    app.build()
    result = (app.outdir / 'index.html').read_text(encoding='utf8')
    assert '<link rel="stylesheet" type="text/css" href="_static/default.css" />' in result
    assert ('<link rel="stylesheet" type="text/css" href="_static/alabaster.css" />'
            not in result)


@pytest.mark.sphinx('html', testroot='basic')
def test_html_sidebar(app, status, warning):
    ctx = {}

    # default for alabaster
    app.build(force_all=True)
    result = (app.outdir / 'index.html').read_text(encoding='utf8')
    assert ('<div class="sphinxsidebar" role="navigation" '
            'aria-label="main navigation">' in result)
    assert '<h1 class="logo"><a href="#">Python</a></h1>' in result
    assert '<h3>Navigation</h3>' in result
    assert '<h3>Related Topics</h3>' in result
    assert '<h3 id="searchlabel">Quick search</h3>' in result

    app.builder.add_sidebars('index', ctx)
    assert ctx['sidebars'] == ['about.html', 'navigation.html', 'relations.html',
                               'searchbox.html', 'donate.html']

    # only relations.html
    app.config.html_sidebars = {'**': ['relations.html']}
    app.build(force_all=True)
    result = (app.outdir / 'index.html').read_text(encoding='utf8')
    assert ('<div class="sphinxsidebar" role="navigation" '
            'aria-label="main navigation">' in result)
    assert '<h1 class="logo"><a href="#">Python</a></h1>' not in result
    assert '<h3>Navigation</h3>' not in result
    assert '<h3>Related Topics</h3>' in result
    assert '<h3 id="searchlabel">Quick search</h3>' not in result

    app.builder.add_sidebars('index', ctx)
    assert ctx['sidebars'] == ['relations.html']

    # no sidebars
    app.config.html_sidebars = {'**': []}
    app.build(force_all=True)
    result = (app.outdir / 'index.html').read_text(encoding='utf8')
    assert ('<div class="sphinxsidebar" role="navigation" '
            'aria-label="main navigation">' not in result)
    assert '<h1 class="logo"><a href="#">Python</a></h1>' not in result
    assert '<h3>Navigation</h3>' not in result
    assert '<h3>Related Topics</h3>' not in result
    assert '<h3 id="searchlabel">Quick search</h3>' not in result

    app.builder.add_sidebars('index', ctx)
    assert ctx['sidebars'] == []


@pytest.mark.parametrize(("fname", "expect"), [
    ('index.html', (".//h1/em/a[@href='https://example.com/cp.1']", '', True)),
    ('index.html', (".//em/a[@href='https://example.com/man.1']", '', True)),
    ('index.html', (".//em/a[@href='https://example.com/ls.1']", '', True)),
    ('index.html', (".//em/a[@href='https://example.com/sphinx.']", '', True)),
])
@pytest.mark.sphinx('html', testroot='manpage_url', confoverrides={
    'manpages_url': 'https://example.com/{page}.{section}'})
@pytest.mark.test_params(shared_result='test_build_html_manpage_url')
def test_html_manpage(app, cached_etree_parse, fname, expect):
    app.build()
    check_xpath(cached_etree_parse(app.outdir / fname), fname, *expect)


@pytest.mark.sphinx('html', testroot='toctree-glob',
                    confoverrides={'html_baseurl': 'https://example.com/'})
def test_html_baseurl(app, status, warning):
    app.build()

    result = (app.outdir / 'index.html').read_text(encoding='utf8')
    assert '<link rel="canonical" href="https://example.com/index.html" />' in result

    result = (app.outdir / 'qux' / 'index.html').read_text(encoding='utf8')
    assert '<link rel="canonical" href="https://example.com/qux/index.html" />' in result


@pytest.mark.sphinx('html', testroot='toctree-glob',
                    confoverrides={'html_baseurl': 'https://example.com/subdir',
                                   'html_file_suffix': '.htm'})
def test_html_baseurl_and_html_file_suffix(app, status, warning):
    app.build()

    result = (app.outdir / 'index.htm').read_text(encoding='utf8')
    assert '<link rel="canonical" href="https://example.com/subdir/index.htm" />' in result

    result = (app.outdir / 'qux' / 'index.htm').read_text(encoding='utf8')
    assert '<link rel="canonical" href="https://example.com/subdir/qux/index.htm" />' in result


@pytest.mark.sphinx(testroot='basic', srcdir='validate_html_extra_path')
def test_validate_html_extra_path(app):
    (app.confdir / '_static').mkdir(parents=True, exist_ok=True)
    app.config.html_extra_path = [
        '/path/to/not_found',       # not found
        '_static',
        app.outdir,                 # outdir
        app.outdir / '_static',     # inside outdir
    ]
    with pytest.warns(RemovedInSphinx80Warning, match='Use "pathlib.Path" or "os.fspath" instead'):
        validate_html_extra_path(app, app.config)
    assert app.config.html_extra_path == ['_static']


@pytest.mark.sphinx(testroot='basic', srcdir='validate_html_static_path')
def test_validate_html_static_path(app):
    (app.confdir / '_static').mkdir(parents=True, exist_ok=True)
    app.config.html_static_path = [
        '/path/to/not_found',       # not found
        '_static',
        app.outdir,                 # outdir
        app.outdir / '_static',     # inside outdir
    ]
    with pytest.warns(RemovedInSphinx80Warning, match='Use "pathlib.Path" or "os.fspath" instead'):
        validate_html_static_path(app, app.config)
    assert app.config.html_static_path == ['_static']


@pytest.mark.sphinx('html', testroot='basic',
                    confoverrides={'html_permalinks': False})
def test_html_permalink_disable(app):
    app.build()
    content = (app.outdir / 'index.html').read_text(encoding='utf8')

    assert '<h1>The basic Sphinx documentation for testing</h1>' in content


@pytest.mark.sphinx('html', testroot='basic',
                    confoverrides={'html_permalinks_icon': '<span>[PERMALINK]</span>'})
def test_html_permalink_icon(app):
    app.build()
    content = (app.outdir / 'index.html').read_text(encoding='utf8')

    assert ('<h1>The basic Sphinx documentation for testing<a class="headerlink" '
            'href="#the-basic-sphinx-documentation-for-testing" '
            'title="Link to this heading"><span>[PERMALINK]</span></a></h1>' in content)


@pytest.mark.sphinx('html', testroot='html_signaturereturn_icon')
def test_html_signaturereturn_icon(app):
    app.build()
    content = (app.outdir / 'index.html').read_text(encoding='utf8')

    assert ('<span class="sig-return-icon">&#x2192;</span>' in content)


@pytest.mark.sphinx('html', testroot='root', srcdir=os.urandom(4).hex())
def test_html_remove_sources_before_write_gh_issue_10786(app, warning):
    # see:  https://github.com/sphinx-doc/sphinx/issues/10786
    target = app.srcdir / 'img.png'

    def handler(app):
        assert target.exists()
        target.unlink()
        return []

    app.connect('html-collect-pages', handler)
    assert target.exists()
    app.build()
    assert not target.exists()

    ws = strip_colors(warning.getvalue()).splitlines()
    assert len(ws) >= 1

    file = os.fsdecode(target)
    assert f'WARNING: cannot copy image file {file!r}: {file!s} does not exist' == ws[-1]



"""Format colored console output."""

from __future__ import annotations

import os
import re
import shutil
import sys
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Final

    # fmt: off
    def reset(text: str) -> str: ...  # NoQA: E704
    def bold(text: str) -> str: ...  # NoQA: E704
    def faint(text: str) -> str: ...  # NoQA: E704
    def standout(text: str) -> str: ...  # NoQA: E704
    def underline(text: str) -> str: ...  # NoQA: E704
    def blink(text: str) -> str: ...  # NoQA: E704

    def black(text: str) -> str: ...  # NoQA: E704
    def white(text: str) -> str: ...  # NoQA: E704
    def red(text: str) -> str: ...  # NoQA: E704
    def green(text: str) -> str: ...  # NoQA: E704
    def yellow(text: str) -> str: ...  # NoQA: E704
    def blue(text: str) -> str: ...  # NoQA: E704
    def fuchsia(text: str) -> str: ...  # NoQA: E704
    def teal(text: str) -> str: ...  # NoQA: E704

    def darkgray(text: str) -> str: ...  # NoQA: E704
    def lightgray(text: str) -> str: ...  # NoQA: E704
    def darkred(text: str) -> str: ...  # NoQA: E704
    def darkgreen(text: str) -> str: ...  # NoQA: E704
    def brown(text: str) -> str: ...  # NoQA: E704
    def darkblue(text: str) -> str: ...  # NoQA: E704
    def purple(text: str) -> str: ...  # NoQA: E704
    def turquoise(text: str) -> str: ...  # NoQA: E704
    # fmt: on

try:
    # check if colorama is installed to support color on Windows
    import colorama
except ImportError:
    colorama = None

_CSI: Final[str] = re.escape('\x1b[')  # 'ESC [': Control Sequence Introducer

# Pattern matching ANSI control sequences containing colors.
_ansi_color_re: Final[re.Pattern[str]] = re.compile(r'\x1b\[(?:\d+;){0,2}\d*m')

_ansi_re: Final[re.Pattern[str]] = re.compile(
    _CSI
    + r"""
    (?:
      (?:\d+;){0,2}\d*m     # ANSI color code    ('m' is equivalent to '0m')
    |
      [012]?K               # ANSI Erase in Line ('K' is equivalent to '0K')
    )""",
    re.VERBOSE | re.ASCII,
)
"""Pattern matching ANSI CSI colors (SGR) and erase line (EL) sequences.

See :func:`strip_escape_sequences` for details.
"""

codes: dict[str, str] = {}


def terminal_safe(s: str) -> str:
    """Safely encode a string for printing to the terminal."""
    return s.encode('ascii', 'backslashreplace').decode('ascii')


def get_terminal_width() -> int:
    """Return the width of the terminal in columns."""
    return shutil.get_terminal_size().columns - 1


_tw: int = get_terminal_width()


def term_width_line(text: str) -> str:
    if not codes:
        # if no coloring, don't output fancy backspaces
        return text + '\n'
    else:
        # codes are not displayed, this must be taken into account
        return text.ljust(_tw + len(text) - len(strip_escape_sequences(text))) + '\r'


def color_terminal() -> bool:
    if 'NO_COLOR' in os.environ:
        return False
    if sys.platform == 'win32' and colorama is not None:
        colorama.init()
        return True
    if 'FORCE_COLOR' in os.environ:
        return True
    if not hasattr(sys.stdout, 'isatty'):
        return False
    if not sys.stdout.isatty():
        return False
    if 'COLORTERM' in os.environ:
        return True
    term = os.environ.get('TERM', 'dumb').lower()
    return term in ('xterm', 'linux') or 'color' in term


def nocolor() -> None:
    if sys.platform == 'win32' and colorama is not None:
        colorama.deinit()
    codes.clear()


def coloron() -> None:
    codes.update(_orig_codes)


def colorize(name: str, text: str, input_mode: bool = False) -> str:
    def escseq(name: str) -> str:
        # Wrap escape sequence with ``\1`` and ``\2`` to let readline know
        # it is non-printable characters
        # ref: https://tiswww.case.edu/php/chet/readline/readline.html
        #
        # Note: This hack does not work well in Windows (see #5059)
        escape = codes.get(name, '')
        if input_mode and escape and sys.platform != 'win32':
            return '\1' + escape + '\2'
        else:
            return escape

    return escseq(name) + text + escseq('reset')


def strip_colors(s: str) -> str:
    """Remove the ANSI color codes in a string *s*.

    .. caution::

       This function is not meant to be used in production and should only
       be used for testing Sphinx's output messages.

    .. seealso:: :func:`strip_escape_sequences`
    """
    return _ansi_color_re.sub('', s)


def strip_escape_sequences(text: str, /) -> str:
    r"""Remove the ANSI CSI colors and "erase in line" sequences.

    Other `escape sequences `__ (e.g., VT100-specific functions) are not
    supported and only control sequences *natively* known to Sphinx (i.e.,
    colors declared in this module and "erase entire line" (``'\x1b[2K'``))
    are eliminated by this function.

    .. caution::

       This function is not meant to be used in production and should only
       be used for testing Sphinx's output messages that were not tempered
       with by third-party extensions.

    .. versionadded:: 7.3

       This function is added as an *experimental* feature.

    __ https://en.wikipedia.org/wiki/ANSI_escape_code
    """
    return _ansi_re.sub('', text)


def create_color_func(name: str) -> None:
    def inner(text: str) -> str:
        return colorize(name, text)

    globals()[name] = inner


_attrs = {
    'reset': '39;49;00m',
    'bold': '01m',
    'faint': '02m',
    'standout': '03m',
    'underline': '04m',
    'blink': '05m',
}

for __name, __value in _attrs.items():
    codes[__name] = '\x1b[' + __value

_colors = [
    ('black', 'darkgray'),
    ('darkred', 'red'),
    ('darkgreen', 'green'),
    ('brown', 'yellow'),
    ('darkblue', 'blue'),
    ('purple', 'fuchsia'),
    ('turquoise', 'teal'),
    ('lightgray', 'white'),
]

for __i, (__dark, __light) in enumerate(_colors, 30):
    codes[__dark] = '\x1b[%im' % __i
    codes[__light] = '\x1b[%im' % (__i + 60)

_orig_codes = codes.copy()

for _name in codes:
    create_color_func(_name)



class Foo:
    """The foo."""
    def do(
        self,
        *,
        keyword_paramtype,
        keyword_kwtype,
        kwarg_paramtype,
        kwarg_kwtype,
        kwparam_paramtype,
        kwparam_kwtype,
    ):
        """Some method.

        :keyword keyword_paramtype: some param
        :paramtype keyword_paramtype: list[int]
        :keyword keyword_kwtype: some param
        :kwtype keyword_kwtype: list[int]
        :kwarg kwarg_paramtype: some param
        :paramtype kwarg_paramtype: list[int]
        :kwarg kwarg_kwtype: some param
        :kwtype kwarg_kwtype: list[int]
        :kwparam kwparam_paramtype: some param
        :paramtype kwparam_paramtype: list[int]
        :kwparam kwparam_kwtype: some param
        :kwtype kwparam_kwtype: list[int]
        """



"""Test sphinx.ext.githubpages extension."""

import pytest


@pytest.mark.sphinx('html', testroot='ext-githubpages')
def test_githubpages(app, status, warning):
    app.build(force_all=True)
    assert (app.outdir / '.nojekyll').exists()
    assert not (app.outdir / 'CNAME').exists()


@pytest.mark.sphinx('html', testroot='ext-githubpages',
                    confoverrides={'html_baseurl': 'https://sphinx-doc.github.io'})
def test_no_cname_for_github_io_domain(app, status, warning):
    app.build(force_all=True)
    assert (app.outdir / '.nojekyll').exists()
    assert not (app.outdir / 'CNAME').exists()


@pytest.mark.sphinx('html', testroot='ext-githubpages',
                    confoverrides={'html_baseurl': 'https://sphinx-doc.org'})
def test_cname_for_custom_domain(app, status, warning):
    app.build(force_all=True)
    assert (app.outdir / '.nojekyll').exists()
    assert (app.outdir / 'CNAME').read_text(encoding='utf8') == 'sphinx-doc.org'



import pytest

from sphinx.errors import ConfigError


@pytest.mark.sphinx('html', testroot='basic')
def test_default_html_math_renderer(app, status, warning):
    assert app.builder.math_renderer_name == 'mathjax'


@pytest.mark.sphinx('html', testroot='basic',
                    confoverrides={'extensions': ['sphinx.ext.mathjax']})
def test_html_math_renderer_is_mathjax(app, status, warning):
    assert app.builder.math_renderer_name == 'mathjax'


@pytest.mark.sphinx('html', testroot='basic',
                    confoverrides={'extensions': ['sphinx.ext.imgmath']})
def test_html_math_renderer_is_imgmath(app, status, warning):
    assert app.builder.math_renderer_name == 'imgmath'


@pytest.mark.sphinx('html', testroot='basic',
                    confoverrides={'extensions': ['sphinxcontrib.jsmath',
                                                  'sphinx.ext.imgmath']})
def test_html_math_renderer_is_duplicated(make_app, app_params):
    args, kwargs = app_params
    with pytest.raises(
        ConfigError,
        match='Many math_renderers are registered. But no math_renderer is selected.',
    ):
        make_app(*args, **kwargs)


@pytest.mark.sphinx('html', testroot='basic',
                    confoverrides={'extensions': ['sphinx.ext.imgmath',
                                                  'sphinx.ext.mathjax']})
def test_html_math_renderer_is_duplicated2(app, status, warning):
    # case of both mathjax and another math_renderer is loaded
    assert app.builder.math_renderer_name == 'imgmath'  # The another one is chosen


@pytest.mark.sphinx('html', testroot='basic',
                    confoverrides={'extensions': ['sphinxcontrib.jsmath',
                                                  'sphinx.ext.imgmath'],
                                   'html_math_renderer': 'imgmath'})
def test_html_math_renderer_is_chosen(app, status, warning):
    assert app.builder.math_renderer_name == 'imgmath'


@pytest.mark.sphinx('html', testroot='basic',
                    confoverrides={'extensions': ['sphinxcontrib.jsmath',
                                                  'sphinx.ext.mathjax'],
                                   'html_math_renderer': 'imgmath'})
def test_html_math_renderer_is_mismatched(make_app, app_params):
    args, kwargs = app_params
    with pytest.raises(ConfigError, match="Unknown math_renderer 'imgmath' is given."):
        make_app(*args, **kwargs)



import pytest


@pytest.mark.sphinx('html', testroot='reST-code-block',
                    confoverrides={'html_codeblock_linenos_style': 'table'})
def test_html_codeblock_linenos_style_table(app):
    app.build()
    content = (app.outdir / 'index.html').read_text(encoding='utf8')

    assert ('<div class="linenodiv"><pre><span class="normal">1</span>\n'
            '<span class="normal">2</span>\n'
            '<span class="normal">3</span>\n'
            '<span class="normal">4</span></pre></div>') in content


@pytest.mark.sphinx('html', testroot='reST-code-block',
                    confoverrides={'html_codeblock_linenos_style': 'inline'})
def test_html_codeblock_linenos_style_inline(app):
    app.build()
    content = (app.outdir / 'index.html').read_text(encoding='utf8')

    assert '<span class="linenos">1</span>' in content


@pytest.mark.sphinx('html', testroot='reST-code-role')
def test_html_code_role(app):
    app.build()
    content = (app.outdir / 'index.html').read_text(encoding='utf8')

    common_content = (
        '<span class="k">def</span> <span class="nf">foo</span>'
        '<span class="p">(</span>'
        '<span class="mi">1</span> '
        '<span class="o">+</span> '
        '<span class="mi">2</span> '
        '<span class="o">+</span> '
        '<span class="kc">None</span> '
        '<span class="o">+</span> '
        '<span class="s2">&quot;abc&quot;</span>'
        '<span class="p">):</span> '
        '<span class="k">pass</span>')
    assert ('<p>Inline <code class="code highlight python docutils literal highlight-python">' +
            common_content + '</code> code block</p>') in content
    assert ('<div class="highlight-python notranslate">' +
            '<div class="highlight"><pre><span></span>' +
            common_content) in content



"""Tests the Python Domain"""

from __future__ import annotations

from docutils import nodes

from sphinx import addnodes
from sphinx.addnodes import (
    desc,
    desc_addname,
    desc_annotation,
    desc_content,
    desc_name,
    desc_parameterlist,
    desc_sig_punctuation,
    desc_sig_space,
    desc_signature,
    pending_xref,
)
from sphinx.testing import restructuredtext
from sphinx.testing.util import assert_node


def test_pyexception_signature(app):
    text = ".. py:exception:: builtins.IOError"
    doctree = restructuredtext.parse(app, text)
    assert_node(doctree, (addnodes.index,
                          [desc, ([desc_signature, ([desc_annotation, ('exception', desc_sig_space)],
                                                    [desc_addname, "builtins."],
                                                    [desc_name, "IOError"])],
                                  desc_content)]))
    assert_node(doctree[1], desc, desctype="exception",
                domain="py", objtype="exception", no_index=False)


def test_pydata_signature(app):
    text = (".. py:data:: version\n"
            "   :type: int\n"
            "   :value: 1\n")
    doctree = restructuredtext.parse(app, text)
    assert_node(doctree, (addnodes.index,
                          [desc, ([desc_signature, ([desc_name, "version"],
                                                    [desc_annotation, ([desc_sig_punctuation, ':'],
                                                                       desc_sig_space,
                                                                       [pending_xref, "int"])],
                                                    [desc_annotation, (
                                                        desc_sig_space,
                                                        [desc_sig_punctuation, '='],
                                                        desc_sig_space,
                                                        "1")],
                                                    )],
                                  desc_content)]))
    assert_node(doctree[1], addnodes.desc, desctype="data",
                domain="py", objtype="data", no_index=False)


def test_pydata_signature_old(app):
    text = (".. py:data:: version\n"
            "   :annotation: = 1\n")
    doctree = restructuredtext.parse(app, text)
    assert_node(doctree, (addnodes.index,
                          [desc, ([desc_signature, ([desc_name, "version"],
                                                    [desc_annotation, (desc_sig_space,
                                                                       "= 1")])],
                                  desc_content)]))
    assert_node(doctree[1], addnodes.desc, desctype="data",
                domain="py", objtype="data", no_index=False)


def test_pydata_with_union_type_operator(app):
    text = (".. py:data:: version\n"
            "   :type: int | str")
    doctree = restructuredtext.parse(app, text)
    assert_node(doctree[1][0],
                ([desc_name, "version"],
                 [desc_annotation, ([desc_sig_punctuation, ':'],
                                    desc_sig_space,
                                    [pending_xref, "int"],
                                    desc_sig_space,
                                    [desc_sig_punctuation, "|"],
                                    desc_sig_space,
                                    [pending_xref, "str"])]))


def test_pyobject_prefix(app):
    text = (".. py:class:: Foo\n"
            "\n"
            "   .. py:method:: Foo.say\n"
            "   .. py:method:: FooBar.say")
    doctree = restructuredtext.parse(app, text)
    assert_node(doctree, (addnodes.index,
                          [desc, ([desc_signature, ([desc_annotation, ('class', desc_sig_space)],
                                                    [desc_name, "Foo"])],
                                  [desc_content, (addnodes.index,
                                                  desc,
                                                  addnodes.index,
                                                  desc)])]))
    assert doctree[1][1][1].astext().strip() == 'say()'           # prefix is stripped
    assert doctree[1][1][3].astext().strip() == 'FooBar.say()'    # not stripped


def test_pydata(app):
    text = (".. py:module:: example\n"
            ".. py:data:: var\n"
            "   :type: int\n")
    domain = app.env.get_domain('py')
    doctree = restructuredtext.parse(app, text)
    assert_node(doctree, (addnodes.index,
                          addnodes.index,
                          nodes.target,
                          [desc, ([desc_signature, ([desc_addname, "example."],
                                                    [desc_name, "var"],
                                                    [desc_annotation, ([desc_sig_punctuation, ':'],
                                                                       desc_sig_space,
                                                                       [pending_xref, "int"])])],
                                  [desc_content, ()])]))
    assert_node(doctree[3][0][2][2], pending_xref, **{"py:module": "example"})
    assert 'example.var' in domain.objects
    assert domain.objects['example.var'] == ('index', 'example.var', 'data', False)


def test_pyclass_options(app):
    text = (".. py:class:: Class1\n"
            ".. py:class:: Class2\n"
            "   :final:\n")
    domain = app.env.get_domain('py')
    doctree = restructuredtext.parse(app, text)
    assert_node(doctree, (addnodes.index,
                          [desc, ([desc_signature, ([desc_annotation, ("class", desc_sig_space)],
                                                    [desc_name, "Class1"])],
                                  [desc_content, ()])],
                          addnodes.index,
                          [desc, ([desc_signature, ([desc_annotation, ("final",
                                                                       desc_sig_space,
                                                                       "class",
                                                                       desc_sig_space)],
                                                    [desc_name, "Class2"])],
                                  [desc_content, ()])]))

    # class
    assert_node(doctree[0], addnodes.index,
                entries=[('single', 'Class1 (built-in class)', 'Class1', '', None)])
    assert 'Class1' in domain.objects
    assert domain.objects['Class1'] == ('index', 'Class1', 'class', False)

    # :final:
    assert_node(doctree[2], addnodes.index,
                entries=[('single', 'Class2 (built-in class)', 'Class2', '', None)])
    assert 'Class2' in domain.objects
    assert domain.objects['Class2'] == ('index', 'Class2', 'class', False)


def test_pymethod_options(app):
    text = (".. py:class:: Class\n"
            "\n"
            "   .. py:method:: meth1\n"
            "   .. py:method:: meth2\n"
            "      :classmethod:\n"
            "   .. py:method:: meth3\n"
            "      :staticmethod:\n"
            "   .. py:method:: meth4\n"
            "      :async:\n"
            "   .. py:method:: meth5\n"
            "      :abstractmethod:\n"
            "   .. py:method:: meth6\n"
            "      :final:\n")
    domain = app.env.get_domain('py')
    doctree = restructuredtext.parse(app, text)
    assert_node(doctree, (addnodes.index,
                          [desc, ([desc_signature, ([desc_annotation, ("class", desc_sig_space)],
                                                    [desc_name, "Class"])],
                                  [desc_content, (addnodes.index,
                                                  desc,
                                                  addnodes.index,
                                                  desc,
                                                  addnodes.index,
                                                  desc,
                                                  addnodes.index,
                                                  desc,
                                                  addnodes.index,
                                                  desc,
                                                  addnodes.index,
                                                  desc)])]))

    # method
    assert_node(doctree[1][1][0], addnodes.index,
                entries=[('single', 'meth1() (Class method)', 'Class.meth1', '', None)])
    assert_node(doctree[1][1][1], ([desc_signature, ([desc_name, "meth1"],
                                                     [desc_parameterlist, ()])],
                                   [desc_content, ()]))
    assert 'Class.meth1' in domain.objects
    assert domain.objects['Class.meth1'] == ('index', 'Class.meth1', 'method', False)

    # :classmethod:
    assert_node(doctree[1][1][2], addnodes.index,
                entries=[('single', 'meth2() (Class class method)', 'Class.meth2', '', None)])
    assert_node(doctree[1][1][3], ([desc_signature, ([desc_annotation, ("classmethod", desc_sig_space)],
                                                     [desc_name, "meth2"],
                                                     [desc_parameterlist, ()])],
                                   [desc_content, ()]))
    assert 'Class.meth2' in domain.objects
    assert domain.objects['Class.meth2'] == ('index', 'Class.meth2', 'method', False)

    # :staticmethod:
    assert_node(doctree[1][1][4], addnodes.index,
                entries=[('single', 'meth3() (Class static method)', 'Class.meth3', '', None)])
    assert_node(doctree[1][1][5], ([desc_signature, ([desc_annotation, ("static", desc_sig_space)],
                                                     [desc_name, "meth3"],
                                                     [desc_parameterlist, ()])],
                                   [desc_content, ()]))
    assert 'Class.meth3' in domain.objects
    assert domain.objects['Class.meth3'] == ('index', 'Class.meth3', 'method', False)

    # :async:
    assert_node(doctree[1][1][6], addnodes.index,
                entries=[('single', 'meth4() (Class method)', 'Class.meth4', '', None)])
    assert_node(doctree[1][1][7], ([desc_signature, ([desc_annotation, ("async", desc_sig_space)],
                                                     [desc_name, "meth4"],
                                                     [desc_parameterlist, ()])],
                                   [desc_content, ()]))
    assert 'Class.meth4' in domain.objects
    assert domain.objects['Class.meth4'] == ('index', 'Class.meth4', 'method', False)

    # :abstractmethod:
    assert_node(doctree[1][1][8], addnodes.index,
                entries=[('single', 'meth5() (Class method)', 'Class.meth5', '', None)])
    assert_node(doctree[1][1][9], ([desc_signature, ([desc_annotation, ("abstract", desc_sig_space)],
                                                     [desc_name, "meth5"],
                                                     [desc_parameterlist, ()])],
                                   [desc_content, ()]))
    assert 'Class.meth5' in domain.objects
    assert domain.objects['Class.meth5'] == ('index', 'Class.meth5', 'method', False)

    # :final:
    assert_node(doctree[1][1][10], addnodes.index,
                entries=[('single', 'meth6() (Class method)', 'Class.meth6', '', None)])
    assert_node(doctree[1][1][11], ([desc_signature, ([desc_annotation, ("final", desc_sig_space)],
                                                      [desc_name, "meth6"],
                                                      [desc_parameterlist, ()])],
                                    [desc_content, ()]))
    assert 'Class.meth6' in domain.objects
    assert domain.objects['Class.meth6'] == ('index', 'Class.meth6', 'method', False)


def test_pyclassmethod(app):
    text = (".. py:class:: Class\n"
            "\n"
            "   .. py:classmethod:: meth\n")
    domain = app.env.get_domain('py')
    doctree = restructuredtext.parse(app, text)
    assert_node(doctree, (addnodes.index,
                          [desc, ([desc_signature, ([desc_annotation, ("class", desc_sig_space)],
                                                    [desc_name, "Class"])],
                                  [desc_content, (addnodes.index,
                                                  desc)])]))
    assert_node(doctree[1][1][0], addnodes.index,
                entries=[('single', 'meth() (Class class method)', 'Class.meth', '', None)])
    assert_node(doctree[1][1][1], ([desc_signature, ([desc_annotation, ("classmethod", desc_sig_space)],
                                                     [desc_name, "meth"],
                                                     [desc_parameterlist, ()])],
                                   [desc_content, ()]))
    assert 'Class.meth' in domain.objects
    assert domain.objects['Class.meth'] == ('index', 'Class.meth', 'method', False)


def test_pystaticmethod(app):
    text = (".. py:class:: Class\n"
            "\n"
            "   .. py:staticmethod:: meth\n")
    domain = app.env.get_domain('py')
    doctree = restructuredtext.parse(app, text)
    assert_node(doctree, (addnodes.index,
                          [desc, ([desc_signature, ([desc_annotation, ("class", desc_sig_space)],
                                                    [desc_name, "Class"])],
                                  [desc_content, (addnodes.index,
                                                  desc)])]))
    assert_node(doctree[1][1][0], addnodes.index,
                entries=[('single', 'meth() (Class static method)', 'Class.meth', '', None)])
    assert_node(doctree[1][1][1], ([desc_signature, ([desc_annotation, ("static", desc_sig_space)],
                                                     [desc_name, "meth"],
                                                     [desc_parameterlist, ()])],
                                   [desc_content, ()]))
    assert 'Class.meth' in domain.objects
    assert domain.objects['Class.meth'] == ('index', 'Class.meth', 'method', False)


def test_pyattribute(app):
    text = (".. py:class:: Class\n"
            "\n"
            "   .. py:attribute:: attr\n"
            "      :type: Optional[str]\n"
            "      :value: ''\n")
    domain = app.env.get_domain('py')
    doctree = restructuredtext.parse(app, text)
    assert_node(doctree, (addnodes.index,
                          [desc, ([desc_signature, ([desc_annotation, ("class", desc_sig_space)],
                                                    [desc_name, "Class"])],
                                  [desc_content, (addnodes.index,
                                                  desc)])]))
    assert_node(doctree[1][1][0], addnodes.index,
                entries=[('single', 'attr (Class attribute)', 'Class.attr', '', None)])
    assert_node(doctree[1][1][1], ([desc_signature, ([desc_name, "attr"],
                                                     [desc_annotation, ([desc_sig_punctuation, ':'],
                                                                        desc_sig_space,
                                                                        [pending_xref, "str"],
                                                                        desc_sig_space,
                                                                        [desc_sig_punctuation, "|"],
                                                                        desc_sig_space,
                                                                        [pending_xref, "None"])],
                                                     [desc_annotation, (desc_sig_space,
                                                                        [desc_sig_punctuation, '='],
                                                                        desc_sig_space,
                                                                        "''")],
                                                     )],
                                   [desc_content, ()]))
    assert_node(doctree[1][1][1][0][1][2], pending_xref, **{"py:class": "Class"})
    assert_node(doctree[1][1][1][0][1][6], pending_xref, **{"py:class": "Class"})
    assert 'Class.attr' in domain.objects
    assert domain.objects['Class.attr'] == ('index', 'Class.attr', 'attribute', False)


def test_pyproperty(app):
    text = (".. py:class:: Class\n"
            "\n"
            "   .. py:property:: prop1\n"
            "      :abstractmethod:\n"
            "      :type: str\n"
            "\n"
            "   .. py:property:: prop2\n"
            "      :classmethod:\n"
            "      :type: str\n")
    domain = app.env.get_domain('py')
    doctree = restructuredtext.parse(app, text)
    assert_node(doctree, (addnodes.index,
                          [desc, ([desc_signature, ([desc_annotation, ("class", desc_sig_space)],
                                                    [desc_name, "Class"])],
                                  [desc_content, (addnodes.index,
                                                  desc,
                                                  addnodes.index,
                                                  desc)])]))
    assert_node(doctree[1][1][0], addnodes.index,
                entries=[('single', 'prop1 (Class property)', 'Class.prop1', '', None)])
    assert_node(doctree[1][1][1], ([desc_signature, ([desc_annotation, ("abstract", desc_sig_space,
                                                                        "property", desc_sig_space)],
                                                     [desc_name, "prop1"],
                                                     [desc_annotation, ([desc_sig_punctuation, ':'],
                                                                        desc_sig_space,
                                                                        [pending_xref, "str"])])],
                                   [desc_content, ()]))
    assert_node(doctree[1][1][2], addnodes.index,
                entries=[('single', 'prop2 (Class property)', 'Class.prop2', '', None)])
    assert_node(doctree[1][1][3], ([desc_signature, ([desc_annotation, ("class", desc_sig_space,
                                                                        "property", desc_sig_space)],
                                                     [desc_name, "prop2"],
                                                     [desc_annotation, ([desc_sig_punctuation, ':'],
                                                                        desc_sig_space,
                                                                        [pending_xref, "str"])])],
                                   [desc_content, ()]))
    assert 'Class.prop1' in domain.objects
    assert domain.objects['Class.prop1'] == ('index', 'Class.prop1', 'property', False)
    assert 'Class.prop2' in domain.objects
    assert domain.objects['Class.prop2'] == ('index', 'Class.prop2', 'property', False)


def test_pydecorator_signature(app):
    text = ".. py:decorator:: deco"
    domain = app.env.get_domain('py')
    doctree = restructuredtext.parse(app, text)
    assert_node(doctree, (addnodes.index,
                          [desc, ([desc_signature, ([desc_addname, "@"],
                                                    [desc_name, "deco"])],
                                  desc_content)]))
    assert_node(doctree[1], addnodes.desc, desctype="function",
                domain="py", objtype="function", no_index=False)

    assert 'deco' in domain.objects
    assert domain.objects['deco'] == ('index', 'deco', 'function', False)


def test_pydecoratormethod_signature(app):
    text = ".. py:decoratormethod:: deco"
    domain = app.env.get_domain('py')
    doctree = restructuredtext.parse(app, text)
    assert_node(doctree, (addnodes.index,
                          [desc, ([desc_signature, ([desc_addname, "@"],
                                                    [desc_name, "deco"])],
                                  desc_content)]))
    assert_node(doctree[1], addnodes.desc, desctype="method",
                domain="py", objtype="method", no_index=False)

    assert 'deco' in domain.objects
    assert domain.objects['deco'] == ('index', 'deco', 'method', False)


def test_pycurrentmodule(app):
    text = (".. py:module:: Other\n"
            "\n"
            ".. py:module:: Module\n"
            ".. py:class:: A\n"
            "\n"
            "   .. py:method:: m1\n"
            "   .. py:method:: m2\n"
            "\n"
            ".. py:currentmodule:: Other\n"
            "\n"
            ".. py:class:: B\n"
            "\n"
            "   .. py:method:: m3\n"
            "   .. py:method:: m4\n")
    domain = app.env.get_domain('py')
    doctree = restructuredtext.parse(app, text)
    print(doctree)
    assert_node(
        doctree, (
            addnodes.index,
            addnodes.index,
            addnodes.index,
            nodes.target,
            nodes.target,
            [desc, (
                [desc_signature, (
                    [desc_annotation, ("class", desc_sig_space)],
                    [desc_addname, "Module."],
                    [desc_name, "A"],
                )],
                [desc_content, (
                    addnodes.index,
                    [desc, (
                        [desc_signature, (
                            [desc_name, "m1"],
                            [desc_parameterlist, ()],
                        )],
                        [desc_content, ()],
                    )],
                    addnodes.index,
                    [desc, (
                        [desc_signature, (
                            [desc_name, "m2"],
                            [desc_parameterlist, ()],
                        )],
                        [desc_content, ()],
                    )],
                )],
            )],
            addnodes.index,
            [desc, (
                [desc_signature, (
                    [desc_annotation, ("class", desc_sig_space)],
                    [desc_addname, "Other."],
                    [desc_name, "B"],
                )],
                [desc_content, (
                    addnodes.index,
                    [desc, (
                        [desc_signature, (
                            [desc_name, "m3"],
                            [desc_parameterlist, ()],
                        )],
                        [desc_content, ()],
                    )],
                    addnodes.index,
                    [desc, (
                        [desc_signature, (
                            [desc_name, "m4"],
                            [desc_parameterlist, ()],
                        )],
                        [desc_content, ()],
                    )],
                )],
            )],
        ))
    assert 'Module' in domain.objects
    assert domain.objects['Module'] == ('index', 'module-Module', 'module', False)
    assert 'Other' in domain.objects
    assert domain.objects['Other'] == ('index', 'module-Other', 'module', False)
    assert 'Module.A' in domain.objects
    assert domain.objects['Module.A'] == ('index', 'Module.A', 'class', False)
    assert 'Other.B' in domain.objects
    assert domain.objects['Other.B'] == ('index', 'Other.B', 'class', False)
    assert 'Module.A.m1' in domain.objects
    assert domain.objects['Module.A.m1'] == ('index', 'Module.A.m1', 'method', False)
    assert 'Module.A.m2' in domain.objects
    assert domain.objects['Module.A.m2'] == ('index', 'Module.A.m2', 'method', False)
    assert 'Other.B.m3' in domain.objects
    assert domain.objects['Other.B.m3'] == ('index', 'Other.B.m3', 'method', False)
    assert 'Other.B.m4' in domain.objects
    assert domain.objects['Other.B.m4'] == ('index', 'Other.B.m4', 'method', False)



"""Test the HTML builder and check output against XPath."""

import os
import re

import pytest

from tests.test_builders.xpath_data import FIGURE_CAPTION
from tests.test_builders.xpath_util import check_xpath


@pytest.mark.sphinx('html', testroot='numfig')
@pytest.mark.test_params(shared_result='test_build_html_numfig')
def test_numfig_disabled_warn(app, warning):
    app.build()
    warnings = warning.getvalue()
    assert 'index.rst:47: WARNING: numfig is disabled. :numref: is ignored.' in warnings
    assert 'index.rst:56: WARNING: invalid numfig_format: invalid' not in warnings
    assert 'index.rst:57: WARNING: invalid numfig_format: Fig %s %s' not in warnings


@pytest.mark.parametrize(("fname", "path", "check", "be_found"), [
    ('index.html', FIGURE_CAPTION + "/span[@class='caption-number']", None, True),
    ('index.html', ".//table/caption/span[@class='caption-number']", None, True),
    ('index.html', ".//div[@class='code-block-caption']/"
     "span[@class='caption-number']", None, True),
    ('index.html', ".//li/p/code/span", '^fig1$', True),
    ('index.html', ".//li/p/code/span", '^Figure%s$', True),
    ('index.html', ".//li/p/code/span", '^table-1$', True),
    ('index.html', ".//li/p/code/span", '^Table:%s$', True),
    ('index.html', ".//li/p/code/span", '^CODE_1$', True),
    ('index.html', ".//li/p/code/span", '^Code-%s$', True),
    ('index.html', ".//li/p/a/span", '^Section 1$', True),
    ('index.html', ".//li/p/a/span", '^Section 2.1$', True),
    ('index.html', ".//li/p/code/span", '^Fig.{number}$', True),
    ('index.html', ".//li/p/a/span", '^Sect.1 Foo$', True),

    ('foo.html', FIGURE_CAPTION + "/span[@class='caption-number']", None, True),
    ('foo.html', ".//table/caption/span[@class='caption-number']", None, True),
    ('foo.html', ".//div[@class='code-block-caption']/"
     "span[@class='caption-number']", None, True),

    ('bar.html', FIGURE_CAPTION + "/span[@class='caption-number']", None, True),
    ('bar.html', ".//table/caption/span[@class='caption-number']", None, True),
    ('bar.html', ".//div[@class='code-block-caption']/"
     "span[@class='caption-number']", None, True),

    ('baz.html', FIGURE_CAPTION + "/span[@class='caption-number']", None, True),
    ('baz.html', ".//table/caption/span[@class='caption-number']", None, True),
    ('baz.html', ".//div[@class='code-block-caption']/"
     "span[@class='caption-number']", None, True),
])
@pytest.mark.sphinx('html', testroot='numfig')
@pytest.mark.test_params(shared_result='test_build_html_numfig')
def test_numfig_disabled(app, cached_etree_parse, fname, path, check, be_found):
    app.build()
    check_xpath(cached_etree_parse(app.outdir / fname), fname, path, check, be_found)


@pytest.mark.sphinx(
    'html', testroot='numfig',
    srcdir='test_numfig_without_numbered_toctree_warn',
    confoverrides={'numfig': True})
def test_numfig_without_numbered_toctree_warn(app, warning):
    app.build()
    # remove :numbered: option
    index = (app.srcdir / 'index.rst').read_text(encoding='utf8')
    index = re.sub(':numbered:.*', '', index)
    (app.srcdir / 'index.rst').write_text(index, encoding='utf8')
    app.build()

    warnings = warning.getvalue()
    assert 'index.rst:47: WARNING: numfig is disabled. :numref: is ignored.' not in warnings
    assert 'index.rst:55: WARNING: Failed to create a cross reference. Any number is not assigned: index' in warnings
    assert 'index.rst:56: WARNING: invalid numfig_format: invalid' in warnings
    assert 'index.rst:57: WARNING: invalid numfig_format: Fig %s %s' in warnings


@pytest.mark.parametrize(("fname", "path", "check", "be_found"), [
    ('index.html', FIGURE_CAPTION + "/span[@class='caption-number']", '^Fig. 9 $', True),
    ('index.html', FIGURE_CAPTION + "/span[@class='caption-number']", '^Fig. 10 $', True),
    ('index.html', ".//table/caption/span[@class='caption-number']",
     '^Table 9 $', True),
    ('index.html', ".//table/caption/span[@class='caption-number']",
     '^Table 10 $', True),
    ('index.html', ".//div[@class='code-block-caption']/"
     "span[@class='caption-number']", '^Listing 9 $', True),
    ('index.html', ".//div[@class='code-block-caption']/"
     "span[@class='caption-number']", '^Listing 10 $', True),
    ('index.html', ".//li/p/a/span", '^Fig. 9$', True),
    ('index.html', ".//li/p/a/span", '^Figure6$', True),
    ('index.html', ".//li/p/a/span", '^Table 9$', True),
    ('index.html', ".//li/p/a/span", '^Table:6$', True),
    ('index.html', ".//li/p/a/span", '^Listing 9$', True),
    ('index.html', ".//li/p/a/span", '^Code-6$', True),
    ('index.html', ".//li/p/code/span", '^foo$', True),
    ('index.html', ".//li/p/code/span", '^bar_a$', True),
    ('index.html', ".//li/p/a/span", '^Fig.9 should be Fig.1$', True),
    ('index.html', ".//li/p/code/span", '^Sect.{number}$', True),

    ('foo.html', FIGURE_CAPTION + "/span[@class='caption-number']", '^Fig. 1 $', True),
    ('foo.html', FIGURE_CAPTION + "/span[@class='caption-number']", '^Fig. 2 $', True),
    ('foo.html', FIGURE_CAPTION + "/span[@class='caption-number']", '^Fig. 3 $', True),
    ('foo.html', FIGURE_CAPTION + "/span[@class='caption-number']", '^Fig. 4 $', True),
    ('foo.html', ".//table/caption/span[@class='caption-number']",
     '^Table 1 $', True),
    ('foo.html', ".//table/caption/span[@class='caption-number']",
     '^Table 2 $', True),
    ('foo.html', ".//table/caption/span[@class='caption-number']",
     '^Table 3 $', True),
    ('foo.html', ".//table/caption/span[@class='caption-number']",
     '^Table 4 $', True),
    ('foo.html', ".//div[@class='code-block-caption']/"
     "span[@class='caption-number']", '^Listing 1 $', True),
    ('foo.html', ".//div[@class='code-block-caption']/"
     "span[@class='caption-number']", '^Listing 2 $', True),
    ('foo.html', ".//div[@class='code-block-caption']/"
     "span[@class='caption-number']", '^Listing 3 $', True),
    ('foo.html', ".//div[@class='code-block-caption']/"
     "span[@class='caption-number']", '^Listing 4 $', True),

    ('bar.html', FIGURE_CAPTION + "/span[@class='caption-number']", '^Fig. 5 $', True),
    ('bar.html', FIGURE_CAPTION + "/span[@class='caption-number']", '^Fig. 7 $', True),
    ('bar.html', FIGURE_CAPTION + "/span[@class='caption-number']", '^Fig. 8 $', True),
    ('bar.html', ".//table/caption/span[@class='caption-number']",
     '^Table 5 $', True),
    ('bar.html', ".//table/caption/span[@class='caption-number']",
     '^Table 7 $', True),
    ('bar.html', ".//table/caption/span[@class='caption-number']",
     '^Table 8 $', True),
    ('bar.html', ".//div[@class='code-block-caption']/"
     "span[@class='caption-number']", '^Listing 5 $', True),
    ('bar.html', ".//div[@class='code-block-caption']/"
     "span[@class='caption-number']", '^Listing 7 $', True),
    ('bar.html', ".//div[@class='code-block-caption']/"
     "span[@class='caption-number']", '^Listing 8 $', True),

    ('baz.html', FIGURE_CAPTION + "/span[@class='caption-number']", '^Fig. 6 $', True),
    ('baz.html', ".//table/caption/span[@class='caption-number']",
     '^Table 6 $', True),
    ('baz.html', ".//div[@class='code-block-caption']/"
     "span[@class='caption-number']", '^Listing 6 $', True),
])
@pytest.mark.sphinx(
    'html', testroot='numfig',
    srcdir='test_numfig_without_numbered_toctree',
    confoverrides={'numfig': True})
def test_numfig_without_numbered_toctree(app, cached_etree_parse, fname, path, check, be_found):
    # remove :numbered: option
    index = (app.srcdir / 'index.rst').read_text(encoding='utf8')
    index = re.sub(':numbered:.*', '', index)
    (app.srcdir / 'index.rst').write_text(index, encoding='utf8')

    if not os.listdir(app.outdir):
        app.build()
    check_xpath(cached_etree_parse(app.outdir / fname), fname, path, check, be_found)


@pytest.mark.sphinx('html', testroot='numfig', confoverrides={'numfig': True})
@pytest.mark.test_params(shared_result='test_build_html_numfig_on')
def test_numfig_with_numbered_toctree_warn(app, warning):
    app.build()
    warnings = warning.getvalue()
    assert 'index.rst:47: WARNING: numfig is disabled. :numref: is ignored.' not in warnings
    assert 'index.rst:55: WARNING: Failed to create a cross reference. Any number is not assigned: index' in warnings
    assert 'index.rst:56: WARNING: invalid numfig_format: invalid' in warnings
    assert 'index.rst:57: WARNING: invalid numfig_format: Fig %s %s' in warnings


@pytest.mark.parametrize(("fname", "path", "check", "be_found"), [
    ('index.html', FIGURE_CAPTION + "/span[@class='caption-number']", '^Fig. 1 $', True),
    ('index.html', FIGURE_CAPTION + "/span[@class='caption-number']", '^Fig. 2 $', True),
    ('index.html', ".//table/caption/span[@class='caption-number']",
     '^Table 1 $', True),
    ('index.html', ".//table/caption/span[@class='caption-number']",
     '^Table 2 $', True),
    ('index.html', ".//div[@class='code-block-caption']/"
     "span[@class='caption-number']", '^Listing 1 $', True),
    ('index.html', ".//div[@class='code-block-caption']/"
     "span[@class='caption-number']", '^Listing 2 $', True),
    ('index.html', ".//li/p/a/span", '^Fig. 1$', True),
    ('index.html', ".//li/p/a/span", '^Figure2.2$', True),
    ('index.html', ".//li/p/a/span", '^Table 1$', True),
    ('index.html', ".//li/p/a/span", '^Table:2.2$', True),
    ('index.html', ".//li/p/a/span", '^Listing 1$', True),
    ('index.html', ".//li/p/a/span", '^Code-2.2$', True),
    ('index.html', ".//li/p/a/span", '^Section.1$', True),
    ('index.html', ".//li/p/a/span", '^Section.2.1$', True),
    ('index.html', ".//li/p/a/span", '^Fig.1 should be Fig.1$', True),
    ('index.html', ".//li/p/a/span", '^Sect.1 Foo$', True),

    ('foo.html', FIGURE_CAPTION + "/span[@class='caption-number']", '^Fig. 1.1 $', True),
    ('foo.html', FIGURE_CAPTION + "/span[@class='caption-number']", '^Fig. 1.2 $', True),
    ('foo.html', FIGURE_CAPTION + "/span[@class='caption-number']", '^Fig. 1.3 $', True),
    ('foo.html', FIGURE_CAPTION + "/span[@class='caption-number']", '^Fig. 1.4 $', True),
    ('foo.html', ".//table/caption/span[@class='caption-number']",
     '^Table 1.1 $', True),
    ('foo.html', ".//table/caption/span[@class='caption-number']",
     '^Table 1.2 $', True),
    ('foo.html', ".//table/caption/span[@class='caption-number']",
     '^Table 1.3 $', True),
    ('foo.html', ".//table/caption/span[@class='caption-number']",
     '^Table 1.4 $', True),
    ('foo.html', ".//div[@class='code-block-caption']/"
     "span[@class='caption-number']", '^Listing 1.1 $', True),
    ('foo.html', ".//div[@class='code-block-caption']/"
     "span[@class='caption-number']", '^Listing 1.2 $', True),
    ('foo.html', ".//div[@class='code-block-caption']/"
     "span[@class='caption-number']", '^Listing 1.3 $', True),
    ('foo.html', ".//div[@class='code-block-caption']/"
     "span[@class='caption-number']", '^Listing 1.4 $', True),

    ('bar.html', FIGURE_CAPTION + "/span[@class='caption-number']", '^Fig. 2.1 $', True),
    ('bar.html', FIGURE_CAPTION + "/span[@class='caption-number']", '^Fig. 2.3 $', True),
    ('bar.html', FIGURE_CAPTION + "/span[@class='caption-number']", '^Fig. 2.4 $', True),
    ('bar.html', ".//table/caption/span[@class='caption-number']",
     '^Table 2.1 $', True),
    ('bar.html', ".//table/caption/span[@class='caption-number']",
     '^Table 2.3 $', True),
    ('bar.html', ".//table/caption/span[@class='caption-number']",
     '^Table 2.4 $', True),
    ('bar.html', ".//div[@class='code-block-caption']/"
     "span[@class='caption-number']", '^Listing 2.1 $', True),
    ('bar.html', ".//div[@class='code-block-caption']/"
     "span[@class='caption-number']", '^Listing 2.3 $', True),
    ('bar.html', ".//div[@class='code-block-caption']/"
     "span[@class='caption-number']", '^Listing 2.4 $', True),

    ('baz.html', FIGURE_CAPTION + "/span[@class='caption-number']", '^Fig. 2.2 $', True),
    ('baz.html', ".//table/caption/span[@class='caption-number']",
     '^Table 2.2 $', True),
    ('baz.html', ".//div[@class='code-block-caption']/"
     "span[@class='caption-number']", '^Listing 2.2 $', True),
])
@pytest.mark.sphinx('html', testroot='numfig', confoverrides={'numfig': True})
@pytest.mark.test_params(shared_result='test_build_html_numfig_on')
def test_numfig_with_numbered_toctree(app, cached_etree_parse, fname, path, check, be_found):
    app.build()
    check_xpath(cached_etree_parse(app.outdir / fname), fname, path, check, be_found)


@pytest.mark.sphinx('html', testroot='numfig', confoverrides={
    'numfig': True,
    'numfig_format': {'figure': 'Figure:%s',
                      'table': 'Tab_%s',
                      'code-block': 'Code-%s',
                      'section': 'SECTION-%s'}})
@pytest.mark.test_params(shared_result='test_build_html_numfig_format_warn')
def test_numfig_with_prefix_warn(app, warning):
    app.build()
    warnings = warning.getvalue()
    assert 'index.rst:47: WARNING: numfig is disabled. :numref: is ignored.' not in warnings
    assert 'index.rst:55: WARNING: Failed to create a cross reference. Any number is not assigned: index' in warnings
    assert 'index.rst:56: WARNING: invalid numfig_format: invalid' in warnings
    assert 'index.rst:57: WARNING: invalid numfig_format: Fig %s %s' in warnings


@pytest.mark.parametrize(("fname", "path", "check", "be_found"), [
    ('index.html', FIGURE_CAPTION + "/span[@class='caption-number']", '^Figure:1 $', True),
    ('index.html', FIGURE_CAPTION + "/span[@class='caption-number']", '^Figure:2 $', True),
    ('index.html', ".//table/caption/span[@class='caption-number']",
     '^Tab_1 $', True),
    ('index.html', ".//table/caption/span[@class='caption-number']",
     '^Tab_2 $', True),
    ('index.html', ".//div[@class='code-block-caption']/"
     "span[@class='caption-number']", '^Code-1 $', True),
    ('index.html', ".//div[@class='code-block-caption']/"
     "span[@class='caption-number']", '^Code-2 $', True),
    ('index.html', ".//li/p/a/span", '^Figure:1$', True),
    ('index.html', ".//li/p/a/span", '^Figure2.2$', True),
    ('index.html', ".//li/p/a/span", '^Tab_1$', True),
    ('index.html', ".//li/p/a/span", '^Table:2.2$', True),
    ('index.html', ".//li/p/a/span", '^Code-1$', True),
    ('index.html', ".//li/p/a/span", '^Code-2.2$', True),
    ('index.html', ".//li/p/a/span", '^SECTION-1$', True),
    ('index.html', ".//li/p/a/span", '^SECTION-2.1$', True),
    ('index.html', ".//li/p/a/span", '^Fig.1 should be Fig.1$', True),
    ('index.html', ".//li/p/a/span", '^Sect.1 Foo$', True),

    ('foo.html', FIGURE_CAPTION + "/span[@class='caption-number']", '^Figure:1.1 $', True),
    ('foo.html', FIGURE_CAPTION + "/span[@class='caption-number']", '^Figure:1.2 $', True),
    ('foo.html', FIGURE_CAPTION + "/span[@class='caption-number']", '^Figure:1.3 $', True),
    ('foo.html', FIGURE_CAPTION + "/span[@class='caption-number']", '^Figure:1.4 $', True),
    ('foo.html', ".//table/caption/span[@class='caption-number']",
     '^Tab_1.1 $', True),
    ('foo.html', ".//table/caption/span[@class='caption-number']",
     '^Tab_1.2 $', True),
    ('foo.html', ".//table/caption/span[@class='caption-number']",
     '^Tab_1.3 $', True),
    ('foo.html', ".//table/caption/span[@class='caption-number']",
     '^Tab_1.4 $', True),
    ('foo.html', ".//div[@class='code-block-caption']/"
     "span[@class='caption-number']", '^Code-1.1 $', True),
    ('foo.html', ".//div[@class='code-block-caption']/"
     "span[@class='caption-number']", '^Code-1.2 $', True),
    ('foo.html', ".//div[@class='code-block-caption']/"
     "span[@class='caption-number']", '^Code-1.3 $', True),
    ('foo.html', ".//div[@class='code-block-caption']/"
     "span[@class='caption-number']", '^Code-1.4 $', True),

    ('bar.html', FIGURE_CAPTION + "/span[@class='caption-number']", '^Figure:2.1 $', True),
    ('bar.html', FIGURE_CAPTION + "/span[@class='caption-number']", '^Figure:2.3 $', True),
    ('bar.html', FIGURE_CAPTION + "/span[@class='caption-number']", '^Figure:2.4 $', True),
    ('bar.html', ".//table/caption/span[@class='caption-number']",
     '^Tab_2.1 $', True),
    ('bar.html', ".//table/caption/span[@class='caption-number']",
     '^Tab_2.3 $', True),
    ('bar.html', ".//table/caption/span[@class='caption-number']",
     '^Tab_2.4 $', True),
    ('bar.html', ".//div[@class='code-block-caption']/"
     "span[@class='caption-number']", '^Code-2.1 $', True),
    ('bar.html', ".//div[@class='code-block-caption']/"
     "span[@class='caption-number']", '^Code-2.3 $', True),
    ('bar.html', ".//div[@class='code-block-caption']/"
     "span[@class='caption-number']", '^Code-2.4 $', True),

    ('baz.html', FIGURE_CAPTION + "/span[@class='caption-number']", '^Figure:2.2 $', True),
    ('baz.html', ".//table/caption/span[@class='caption-number']",
     '^Tab_2.2 $', True),
    ('baz.html', ".//div[@class='code-block-caption']/"
     "span[@class='caption-number']", '^Code-2.2 $', True),
])
@pytest.mark.sphinx('html', testroot='numfig',
                    confoverrides={'numfig': True,
                                   'numfig_format': {'figure': 'Figure:%s',
                                                     'table': 'Tab_%s',
                                                     'code-block': 'Code-%s',
                                                     'section': 'SECTION-%s'}})
@pytest.mark.test_params(shared_result='test_build_html_numfig_format_warn')
def test_numfig_with_prefix(app, cached_etree_parse, fname, path, check, be_found):
    app.build()
    check_xpath(cached_etree_parse(app.outdir / fname), fname, path, check, be_found)


@pytest.mark.sphinx('html', testroot='numfig',
                    confoverrides={'numfig': True, 'numfig_secnum_depth': 2})
@pytest.mark.test_params(shared_result='test_build_html_numfig_depth_2')
def test_numfig_with_secnum_depth_warn(app, warning):
    app.build()
    warnings = warning.getvalue()
    assert 'index.rst:47: WARNING: numfig is disabled. :numref: is ignored.' not in warnings
    assert 'index.rst:55: WARNING: Failed to create a cross reference. Any number is not assigned: index' in warnings
    assert 'index.rst:56: WARNING: invalid numfig_format: invalid' in warnings
    assert 'index.rst:57: WARNING: invalid numfig_format: Fig %s %s' in warnings


@pytest.mark.parametrize(("fname", "path", "check", "be_found"), [
    ('index.html', FIGURE_CAPTION + "/span[@class='caption-number']", '^Fig. 1 $', True),
    ('index.html', FIGURE_CAPTION + "/span[@class='caption-number']", '^Fig. 2 $', True),
    ('index.html', ".//table/caption/span[@class='caption-number']",
     '^Table 1 $', True),
    ('index.html', ".//table/caption/span[@class='caption-number']",
     '^Table 2 $', True),
    ('index.html', ".//div[@class='code-block-caption']/"
     "span[@class='caption-number']", '^Listing 1 $', True),
    ('index.html', ".//div[@class='code-block-caption']/"
     "span[@class='caption-number']", '^Listing 2 $', True),
    ('index.html', ".//li/p/a/span", '^Fig. 1$', True),
    ('index.html', ".//li/p/a/span", '^Figure2.1.2$', True),
    ('index.html', ".//li/p/a/span", '^Table 1$', True),
    ('index.html', ".//li/p/a/span", '^Table:2.1.2$', True),
    ('index.html', ".//li/p/a/span", '^Listing 1$', True),
    ('index.html', ".//li/p/a/span", '^Code-2.1.2$', True),
    ('index.html', ".//li/p/a/span", '^Section.1$', True),
    ('index.html', ".//li/p/a/span", '^Section.2.1$', True),
    ('index.html', ".//li/p/a/span", '^Fig.1 should be Fig.1$', True),
    ('index.html', ".//li/p/a/span", '^Sect.1 Foo$', True),

    ('foo.html', FIGURE_CAPTION + "/span[@class='caption-number']", '^Fig. 1.1 $', True),
    ('foo.html', FIGURE_CAPTION + "/span[@class='caption-number']", '^Fig. 1.1.1 $', True),
    ('foo.html', FIGURE_CAPTION + "/span[@class='caption-number']", '^Fig. 1.1.2 $', True),
    ('foo.html', FIGURE_CAPTION + "/span[@class='caption-number']", '^Fig. 1.2.1 $', True),
    ('foo.html', ".//table/caption/span[@class='caption-number']",
     '^Table 1.1 $', True),
    ('foo.html', ".//table/caption/span[@class='caption-number']",
     '^Table 1.1.1 $', True),
    ('foo.html', ".//table/caption/span[@class='caption-number']",
     '^Table 1.1.2 $', True),
    ('foo.html', ".//table/caption/span[@class='caption-number']",
     '^Table 1.2.1 $', True),
    ('foo.html', ".//div[@class='code-block-caption']/"
     "span[@class='caption-number']", '^Listing 1.1 $', True),
    ('foo.html', ".//div[@class='code-block-caption']/"
     "span[@class='caption-number']", '^Listing 1.1.1 $', True),
    ('foo.html', ".//div[@class='code-block-caption']/"
     "span[@class='caption-number']", '^Listing 1.1.2 $', True),
    ('foo.html', ".//div[@class='code-block-caption']/"
     "span[@class='caption-number']", '^Listing 1.2.1 $', True),

    ('bar.html', FIGURE_CAPTION + "/span[@class='caption-number']", '^Fig. 2.1.1 $', True),
    ('bar.html', FIGURE_CAPTION + "/span[@class='caption-number']", '^Fig. 2.1.3 $', True),
    ('bar.html', FIGURE_CAPTION + "/span[@class='caption-number']", '^Fig. 2.2.1 $', True),
    ('bar.html', ".//table/caption/span[@class='caption-number']",
     '^Table 2.1.1 $', True),
    ('bar.html', ".//table/caption/span[@class='caption-number']",
     '^Table 2.1.3 $', True),
    ('bar.html', ".//table/caption/span[@class='caption-number']",
     '^Table 2.2.1 $', True),
    ('bar.html', ".//div[@class='code-block-caption']/"
     "span[@class='caption-number']", '^Listing 2.1.1 $', True),
    ('bar.html', ".//div[@class='code-block-caption']/"
     "span[@class='caption-number']", '^Listing 2.1.3 $', True),
    ('bar.html', ".//div[@class='code-block-caption']/"
     "span[@class='caption-number']", '^Listing 2.2.1 $', True),

    ('baz.html', FIGURE_CAPTION + "/span[@class='caption-number']", '^Fig. 2.1.2 $', True),
    ('baz.html', ".//table/caption/span[@class='caption-number']",
     '^Table 2.1.2 $', True),
    ('baz.html', ".//div[@class='code-block-caption']/"
     "span[@class='caption-number']", '^Listing 2.1.2 $', True),
])
@pytest.mark.sphinx('html', testroot='numfig',
                    confoverrides={'numfig': True,
                                   'numfig_secnum_depth': 2})
@pytest.mark.test_params(shared_result='test_build_html_numfig_depth_2')
def test_numfig_with_secnum_depth(app, cached_etree_parse, fname, path, check, be_found):
    app.build()
    check_xpath(cached_etree_parse(app.outdir / fname), fname, path, check, be_found)


@pytest.mark.parametrize("expect", [
    (FIGURE_CAPTION + "/span[@class='caption-number']", '^Fig. 1 $', True),
    (FIGURE_CAPTION + "/span[@class='caption-number']", '^Fig. 2 $', True),
    (".//table/caption/span[@class='caption-number']",
     '^Table 1 $', True),
    (".//table/caption/span[@class='caption-number']",
     '^Table 2 $', True),
    (".//div[@class='code-block-caption']/"
     "span[@class='caption-number']", '^Listing 1 $', True),
    (".//div[@class='code-block-caption']/"
     "span[@class='caption-number']", '^Listing 2 $', True),
    (".//li/p/a/span", '^Fig. 1$', True),
    (".//li/p/a/span", '^Figure2.2$', True),
    (".//li/p/a/span", '^Table 1$', True),
    (".//li/p/a/span", '^Table:2.2$', True),
    (".//li/p/a/span", '^Listing 1$', True),
    (".//li/p/a/span", '^Code-2.2$', True),
    (".//li/p/a/span", '^Section.1$', True),
    (".//li/p/a/span", '^Section.2.1$', True),
    (".//li/p/a/span", '^Fig.1 should be Fig.1$', True),
    (".//li/p/a/span", '^Sect.1 Foo$', True),
    (FIGURE_CAPTION + "/span[@class='caption-number']", '^Fig. 1.1 $', True),
    (FIGURE_CAPTION + "/span[@class='caption-number']", '^Fig. 1.2 $', True),
    (FIGURE_CAPTION + "/span[@class='caption-number']", '^Fig. 1.3 $', True),
    (FIGURE_CAPTION + "/span[@class='caption-number']", '^Fig. 1.4 $', True),
    (".//table/caption/span[@class='caption-number']",
     '^Table 1.1 $', True),
    (".//table/caption/span[@class='caption-number']",
     '^Table 1.2 $', True),
    (".//table/caption/span[@class='caption-number']",
     '^Table 1.3 $', True),
    (".//table/caption/span[@class='caption-number']",
     '^Table 1.4 $', True),
    (".//div[@class='code-block-caption']/"
     "span[@class='caption-number']", '^Listing 1.1 $', True),
    (".//div[@class='code-block-caption']/"
     "span[@class='caption-number']", '^Listing 1.2 $', True),
    (".//div[@class='code-block-caption']/"
     "span[@class='caption-number']", '^Listing 1.3 $', True),
    (".//div[@class='code-block-caption']/"
     "span[@class='caption-number']", '^Listing 1.4 $', True),
    (FIGURE_CAPTION + "/span[@class='caption-number']", '^Fig. 2.1 $', True),
    (FIGURE_CAPTION + "/span[@class='caption-number']", '^Fig. 2.3 $', True),
    (FIGURE_CAPTION + "/span[@class='caption-number']", '^Fig. 2.4 $', True),
    (".//table/caption/span[@class='caption-number']",
     '^Table 2.1 $', True),
    (".//table/caption/span[@class='caption-number']",
     '^Table 2.3 $', True),
    (".//table/caption/span[@class='caption-number']",
     '^Table 2.4 $', True),
    (".//div[@class='code-block-caption']/"
     "span[@class='caption-number']", '^Listing 2.1 $', True),
    (".//div[@class='code-block-caption']/"
     "span[@class='caption-number']", '^Listing 2.3 $', True),
    (".//div[@class='code-block-caption']/"
     "span[@class='caption-number']", '^Listing 2.4 $', True),
    (FIGURE_CAPTION + "/span[@class='caption-number']", '^Fig. 2.2 $', True),
    (".//table/caption/span[@class='caption-number']",
     '^Table 2.2 $', True),
    (".//div[@class='code-block-caption']/"
     "span[@class='caption-number']", '^Listing 2.2 $', True),
])
@pytest.mark.sphinx('singlehtml', testroot='numfig', confoverrides={'numfig': True})
@pytest.mark.test_params(shared_result='test_build_html_numfig_on')
def test_numfig_with_singlehtml(app, cached_etree_parse, expect):
    app.build()
    check_xpath(cached_etree_parse(app.outdir / 'index.html'), 'index.html', *expect)



from unittest.mock import ANY, call, patch

import pytest


@pytest.mark.sphinx('html', testroot='basic')
def test_html_pygments_style_default(app):
    style = app.builder.highlighter.formatter_args.get('style')
    assert style.__name__ == 'Alabaster'


@pytest.mark.sphinx('html', testroot='basic',
                    confoverrides={'pygments_style': 'sphinx'})
def test_html_pygments_style_manually(app):
    style = app.builder.highlighter.formatter_args.get('style')
    assert style.__name__ == 'SphinxStyle'


@pytest.mark.sphinx('html', testroot='basic',
                    confoverrides={'html_theme': 'classic'})
def test_html_pygments_for_classic_theme(app):
    style = app.builder.highlighter.formatter_args.get('style')
    assert style.__name__ == 'SphinxStyle'


@pytest.mark.sphinx('html', testroot='basic')
def test_html_dark_pygments_style_default(app):
    assert app.builder.dark_highlighter is None


@pytest.mark.sphinx('html', testroot='highlight_options')
def test_highlight_options(app):
    subject = app.builder.highlighter
    with patch.object(subject, 'highlight_block', wraps=subject.highlight_block) as highlight:
        app.build()

        call_args = highlight.call_args_list
        assert len(call_args) == 3
        assert call_args[0] == call(ANY, 'default', force=False, linenos=False,
                                    location=ANY, opts={'default_option': True})
        assert call_args[1] == call(ANY, 'python', force=False, linenos=False,
                                    location=ANY, opts={'python_option': True})
        assert call_args[2] == call(ANY, 'java', force=False, linenos=False,
                                    location=ANY, opts={})


@pytest.mark.sphinx('html', testroot='highlight_options',
                    confoverrides={'highlight_options': {'default_option': True}})
def test_highlight_options_old(app):
    subject = app.builder.highlighter
    with patch.object(subject, 'highlight_block', wraps=subject.highlight_block) as highlight:
        app.build()

        call_args = highlight.call_args_list
        assert len(call_args) == 3
        assert call_args[0] == call(ANY, 'default', force=False, linenos=False,
                                    location=ANY, opts={'default_option': True})
        assert call_args[1] == call(ANY, 'python', force=False, linenos=False,
                                    location=ANY, opts={})
        assert call_args[2] == call(ANY, 'java', force=False, linenos=False,
                                    location=ANY, opts={})



"""Test templating."""

import pytest

from sphinx.ext.autosummary.generate import setup_documenters


@pytest.mark.sphinx('html', testroot='templating')
def test_layout_overloading(make_app, app_params):
    args, kwargs = app_params
    app = make_app(*args, **kwargs)
    setup_documenters(app)
    app.build()

    result = (app.outdir / 'index.html').read_text(encoding='utf8')
    assert '<!-- layout overloading -->' in result


@pytest.mark.sphinx('html', testroot='templating')
def test_autosummary_class_template_overloading(make_app, app_params):
    args, kwargs = app_params
    app = make_app(*args, **kwargs)
    setup_documenters(app)
    app.build()

    result = (app.outdir / 'generated' / 'sphinx.application.TemplateBridge.html').read_text(
        encoding='utf8'
    )
    assert 'autosummary/class.rst method block overloading' in result
    assert 'foobar' not in result


@pytest.mark.sphinx(
    'html',
    testroot='templating',
    confoverrides={'autosummary_context': {'sentence': 'foobar'}},
)
def test_autosummary_context(make_app, app_params):
    args, kwargs = app_params
    app = make_app(*args, **kwargs)
    setup_documenters(app)
    app.build()

    result = (app.outdir / 'generated' / 'sphinx.application.TemplateBridge.html').read_text(
        encoding='utf8'
    )
    assert 'autosummary/class.rst method block overloading' in result
    assert 'foobar' in result



"""Test the HTML builder and check output against XPath."""

import pytest

from tests.test_builders.xpath_util import check_xpath


@pytest.mark.parametrize(("fname", "path", "check", "be_found"), [
    ('index.html', ".//li[@class='toctree-l3']/a", '1.1.1. Foo A1', True),
    ('index.html', ".//li[@class='toctree-l3']/a", '1.2.1. Foo B1', True),
    ('index.html', ".//li[@class='toctree-l3']/a", '2.1.1. Bar A1', False),
    ('index.html', ".//li[@class='toctree-l3']/a", '2.2.1. Bar B1', False),

    ('foo.html', ".//h1", 'Foo', True),
    ('foo.html', ".//h2", 'Foo A', True),
    ('foo.html', ".//h3", 'Foo A1', True),
    ('foo.html', ".//h2", 'Foo B', True),
    ('foo.html', ".//h3", 'Foo B1', True),

    ('foo.html', ".//h1//span[@class='section-number']", '1. ', True),
    ('foo.html', ".//h2//span[@class='section-number']", '1.1. ', True),
    ('foo.html', ".//h3//span[@class='section-number']", '1.1.1. ', True),
    ('foo.html', ".//h2//span[@class='section-number']", '1.2. ', True),
    ('foo.html', ".//h3//span[@class='section-number']", '1.2.1. ', True),

    ('foo.html', ".//div[@class='sphinxsidebarwrapper']//li/a", '1.1. Foo A', True),
    ('foo.html', ".//div[@class='sphinxsidebarwrapper']//li/a", '1.1.1. Foo A1', True),
    ('foo.html', ".//div[@class='sphinxsidebarwrapper']//li/a", '1.2. Foo B', True),
    ('foo.html', ".//div[@class='sphinxsidebarwrapper']//li/a", '1.2.1. Foo B1', True),

    ('bar.html', ".//h1", 'Bar', True),
    ('bar.html', ".//h2", 'Bar A', True),
    ('bar.html', ".//h2", 'Bar B', True),
    ('bar.html', ".//h3", 'Bar B1', True),
    ('bar.html', ".//h1//span[@class='section-number']", '2. ', True),
    ('bar.html', ".//h2//span[@class='section-number']", '2.1. ', True),
    ('bar.html', ".//h2//span[@class='section-number']", '2.2. ', True),
    ('bar.html', ".//h3//span[@class='section-number']", '2.2.1. ', True),
    ('bar.html', ".//div[@class='sphinxsidebarwrapper']//li/a", '2. Bar', True),
    ('bar.html', ".//div[@class='sphinxsidebarwrapper']//li/a", '2.1. Bar A', True),
    ('bar.html', ".//div[@class='sphinxsidebarwrapper']//li/a", '2.2. Bar B', True),
    ('bar.html', ".//div[@class='sphinxsidebarwrapper']//li/a", '2.2.1. Bar B1', False),

    ('baz.html', ".//h1", 'Baz A', True),
    ('baz.html', ".//h1//span[@class='section-number']", '2.1.1. ', True),
])
@pytest.mark.sphinx('html', testroot='tocdepth')
@pytest.mark.test_params(shared_result='test_build_html_tocdepth')
def test_tocdepth(app, cached_etree_parse, fname, path, check, be_found):
    app.build()
    # issue #1251
    check_xpath(cached_etree_parse(app.outdir / fname), fname, path, check, be_found)


@pytest.mark.parametrize("expect", [
    (".//li[@class='toctree-l3']/a", '1.1.1. Foo A1', True),
    (".//li[@class='toctree-l3']/a", '1.2.1. Foo B1', True),
    (".//li[@class='toctree-l3']/a", '2.1.1. Bar A1', False),
    (".//li[@class='toctree-l3']/a", '2.2.1. Bar B1', False),

    # index.rst
    (".//h1", 'test-tocdepth', True),

    # foo.rst
    (".//h2", 'Foo', True),
    (".//h3", 'Foo A', True),
    (".//h4", 'Foo A1', True),
    (".//h3", 'Foo B', True),
    (".//h4", 'Foo B1', True),
    (".//h2//span[@class='section-number']", '1. ', True),
    (".//h3//span[@class='section-number']", '1.1. ', True),
    (".//h4//span[@class='section-number']", '1.1.1. ', True),
    (".//h3//span[@class='section-number']", '1.2. ', True),
    (".//h4//span[@class='section-number']", '1.2.1. ', True),

    # bar.rst
    (".//h2", 'Bar', True),
    (".//h3", 'Bar A', True),
    (".//h3", 'Bar B', True),
    (".//h4", 'Bar B1', True),
    (".//h2//span[@class='section-number']", '2. ', True),
    (".//h3//span[@class='section-number']", '2.1. ', True),
    (".//h3//span[@class='section-number']", '2.2. ', True),
    (".//h4//span[@class='section-number']", '2.2.1. ', True),

    # baz.rst
    (".//h4", 'Baz A', True),
    (".//h4//span[@class='section-number']", '2.1.1. ', True),
])
@pytest.mark.sphinx('singlehtml', testroot='tocdepth')
@pytest.mark.test_params(shared_result='test_build_html_tocdepth')
def test_tocdepth_singlehtml(app, cached_etree_parse, expect):
    app.build()
    check_xpath(cached_etree_parse(app.outdir / 'index.html'), 'index.html', *expect)



"""
    Test case for #11387 corner case involving inherited
    members with type annotations on python 3.9 and earlier
"""

class HasTypeAnnotatedMember:
    inherit_me: int
    """Inherited"""

class NoTypeAnnotation(HasTypeAnnotatedMember):
    a = 1
    """Local"""

class NoTypeAnnotation2(HasTypeAnnotatedMember):
    a = 1
    """Local"""

